#!/usr/bin/python

# Copyright 2019 SchedMD LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import fcntl
import httplib2
import logging
import os
import shlex
import subprocess
import sys
import time
import tempfile

import googleapiclient.discovery

CLUSTER_NAME = '@CLUSTER_NAME@'

PROJECT      = '@PROJECT@'
ZONE         = '@ZONE@'

SCONTROL     = '/apps/slurm/current/bin/scontrol'
SINFO        = '/apps/slurm/current/bin/sinfo'
LOGDIR       = '/apps/slurm/log'

TOT_REQ_CNT = 1000

retry_list = []

# [START start_instances_cb]
def start_instances_cb(request_id, response, exception):
    if exception is not None:
        logging.error("start exception: " + str(exception))
        if "Rate Limit Exceeded" in str(exception):
            retry_list.append(request_id)
        elif "was not found" in str(exception):
            subprocess.Popen(
                shlex.split("/apps/slurm/scripts/resume.py {}"
                            .format(request_id)))
# [END start_instances_cb]


# [START start_instances]
def start_instances(compute, node_list):

    req_cnt = 0
    curr_batch = 0
    batch_list = []
    batch_list.insert(
        curr_batch,
        compute.new_batch_http_request(callback=start_instances_cb))

    for node in node_list:
        if req_cnt >= TOT_REQ_CNT:
            req_cnt = 0
            curr_batch += 1
            batch_list.insert(
                curr_batch,
                compute.new_batch_http_request(callback=start_instances_cb))

        batch_list[curr_batch].add(
            compute.instances().start(project=PROJECT, zone=ZONE,
                                      instance=node),
            request_id=node)
        req_cnt += 1
    try:
        for i, batch in enumerate(batch_list):
            batch.execute()
            if i < (len(batch_list) - 1):
                time.sleep(30)
    except Exception, e:
        logging.exception("error in start batch: " + str(e))

# [END start_instances]

# [START main]
def main():
    compute = googleapiclient.discovery.build('compute', 'v1',
                                              cache_discovery=False)

    try:
        s_nodes = dict()
        cmd = "{} --noheader -N -o'%N,%t' | uniq".format(SINFO)
        nodes = subprocess.check_output(cmd, shell=True)
        if nodes:
            s_nodes = dict(n.split(",") for n in nodes.rstrip().splitlines())

        page_token = ""
        g_nodes = []
        while True:
            resp = compute.instances().list(
                      project=PROJECT, zone=ZONE, pageToken=page_token,
                      filter='name={}-compute*'.format(CLUSTER_NAME)).execute()

            if "items" in resp:
                g_nodes.extend(resp['items'])
            if "nextPageToken" in resp:
                page_token = resp['nextPageToken']
                continue

            break;

        to_down = []
        to_idle = []
        to_start = []
        for s_node, s_state in s_nodes.iteritems():
            g_node = next((item for item in g_nodes
                           if item["name"] == s_node),
                          None)

            if (("~" not in s_state) and ("%" not in s_state)):
                # slurm nodes that aren't in power_save and are stopped in GCP:
                #   mark down in slurm
                #   start them in gcp
                if g_node and (g_node['status'] == "TERMINATED"):
                    to_down.append(s_node)
                    to_start.append(s_node)

                # can't check if the node doesn't exist in GCP while the node
                # is booting because it might not have been created yet by the
                # resume script.
                # This should catch the completing states as well.
                if ((g_node == None) and ("#" not in s_state)):
                    to_down.append(s_node)
            elif (g_node == None):
                # find nodes that are down~ in slurm and don't exist in gcp:
                #   mark idle~
                if (s_state == "down~"):
                    to_idle.append(s_node)
                elif ("%" in s_state):
                    to_idle.append(s_node)
                elif ("comp" in s_state):
                    to_down.append(s_node)

        if len(to_down):
            logging.debug("{} stopped/deleted instances ({})".format(
                len(to_down), ",".join(to_down)))
            logging.debug("{} instances to start ({})".format(
                len(to_start), ",".join(to_start)))

            # write hosts to a file that can be given to get a slurm
            # hostlist. Since the number of hosts could be large.
            tmp_file = tempfile.NamedTemporaryFile(mode='w+t', delete=False)
            tmp_file.writelines("\n".join(to_down))
            tmp_file.close()
            logging.debug("tmp_file = {}".format(tmp_file.name))

            cmd = "{} show hostlist {}".format(SCONTROL, tmp_file.name)
            hostlist = subprocess.check_output(shlex.split(cmd))
            logging.debug("hostlist = {}".format(hostlist))
            os.remove(tmp_file.name)

            cmd = "{} update nodename={} state=down reason='Instance stopped/deleted'".format(
                SCONTROL, hostlist)
            subprocess.call(shlex.split(cmd))

            while True:
                start_instances(compute, to_start)
                if not len(retry_list):
                    break;

                logging.debug("got {} nodes to retry ({})".
                              format(len(retry_list),",".join(retry_list)))
                to_start = list(retry_list)
                del retry_list[:]


        if len(to_idle):
            logging.debug("{} instances to resume ({})".format(
                len(to_idle), ",".join(to_idle)))

            # write hosts to a file that can be given to get a slurm
            # hostlist. Since the number of hosts could be large.
            tmp_file = tempfile.NamedTemporaryFile(mode='w+t', delete=False)
            tmp_file.writelines("\n".join(to_idle))
            tmp_file.close()
            logging.debug("tmp_file = {}".format(tmp_file.name))

            cmd = "{} show hostlist {}".format(SCONTROL, tmp_file.name)
            hostlist = subprocess.check_output(shlex.split(cmd))
            logging.debug("hostlist = {}".format(hostlist))
            os.remove(tmp_file.name)

            cmd = "{} update nodename={} state=resume".format(
                SCONTROL, hostlist)
            subprocess.call(shlex.split(cmd))


    except Exception, e:
        logging.error("failed to sync instances ({})".format(str(e)))

# [END main]


if __name__ == '__main__':
    base = os.path.basename(__file__)
    file_name = os.path.splitext(base)[0]

    # silence module logging
    for logger in logging.Logger.manager.loggerDict:
        logging.getLogger(logger).setLevel(logging.WARNING)

    logging.basicConfig(
        filename="{}/{}.log".format(LOGDIR, file_name),
        format='%(asctime)s %(name)s %(levelname)s: %(message)s',
        level=logging.DEBUG)

    # only run one instance at a time
    pid_file = '/tmp/{}.pid'.format(file_name)
    fp = open(pid_file, 'w')
    try:
        fcntl.lockf(fp, fcntl.LOCK_EX | fcntl.LOCK_NB)
    except IOError:
        sys.exit(0)

    main()



import typing
import os
import re
import io
import sys
import subprocess
import warnings
import binascii
import traceback
import shlex
from .base import AbstractSlurmBackend, AbstractTransport
from ..utils import ArgumentHelper, make_interactive, check_call
from agutil import StdOutAdapter
import pandas as pd
import paramiko

import logging
logging.getLogger('paramiko').setLevel(logging.WARNING)

SSH_AGENT_PATTERN = re.compile(r'SSH_AUTH_SOCK=(.+); export SSH_AUTH_SOCK')

class IgnoreKeyPolicy(paramiko.client.AutoAddPolicy):
    """
    Slight modification of paramiko.client.AutoAddPolicy
    Doesn't save the new key to the file, but accepts it anyways
    """

    def missing_host_key(self, client, hostname, key):
        client._host_keys.add(hostname, key.get_name(), key)

class RemoteTransport(AbstractTransport):
    """
    Transport for working with remote files over ssh
    """
    def __init__(self, client: paramiko.SSHClient):
        """
        Initializes the transport from a given SSH client
        """
        self.client = client
        self.session = None

    def __enter__(self):
        """
        Starts a connection to the remote server
        """
        if self.client._transport is None:
            raise paramiko.SSHException("Client is not connected")
        self.session = self.client.open_sftp()
        return self

    def __exit__(self, *args):
        """
        Closes the connection to the remote server
        """
        self.session.close()
        self.session = None

    def open(self, filename: str, mode: str = 'r', bufsize: int = -1) -> typing.IO:
        """
        Returns a File-Like object open on the slurm cluster
        """
        if self.session is None:
            raise paramiko.SSHException("Transport is not connected")
        handle = self.session.open(filename, mode, bufsize)
        handle.mode = mode
        handle.name = filename
        if 'w' in mode:
            handle.set_pipelined(True)
        return handle

    def listdir(self, path: str) -> typing.List[str]:
        """
        Lists the contents of the requested path
        """
        if self.session is None:
            raise paramiko.SSHException("Transport is not connected")
        return self.session.listdir(path)

    def mkdir(self, path: str):
        """
        Creates the requested directory
        """
        if self.session is None:
            raise paramiko.SSHException("Transport is not connected")
        return self.session.mkdir(path)

    def stat(self, path: str) -> typing.Any:
        """
        Returns stat information
        """
        if self.session is None:
            raise paramiko.SSHException("Transport is not connected")
        return self.session.stat(path)

    def chmod(self, path: str, mode: int):
        """
        Change file permissions
        """
        if self.session is None:
            raise paramiko.SSHException("Transport is not connected")
        self.session.chmod(path, mode)

    def normpath(self, path: str) -> str:
        """
        Returns a normalized path relative to the transport
        """
        if self.session is None:
            raise paramiko.SSHException("Transport is not connected")
        try:
            return self.session.normalize(path)
        except FileNotFoundError:
            if path.startswith('/'):
                return path # Absolute
            # Relative
            return os.path.join(
                self.session.normalize('.'),
                path
            )

    def remove(self, path: str):
        """
        Removes the file at the given path
        """
        if self.session is None:
            raise paramiko.SSHException("Transport is not connected")
        self.session.remove(path)

    def rmdir(self, path: str):
        """
        Removes the directory at the given path
        """
        if self.session is None:
            raise paramiko.SSHException("Transport is not connected")
        self.session.rmdir(path)

    def mklink(self, src: str, dest: str):
        """
        Creates a symlink from dest->src
        """
        if self.session is None:
            raise paramiko.SSHException("Transport is not connected")
        self.session.symlink(src, dest)

    def rename(self, src: str, dest: str):
        """
        Move the file or folder 'src' to 'dest'
        Will overwrite dest if it exists
        """
        if self.session is None:
            raise paramiko.SSHException("Transport is not connected")
        try:
            self.session.posix_rename(src, dest)
        except IOError:
            self.session.rename(src, dest)

class RemoteSlurmBackend(AbstractSlurmBackend):
    """
    SLURM backend for interacting with a remote slurm node
    """

    @staticmethod
    def ssh_agent() -> typing.Optional[str]:
        """
        Ensures proper environment variables are set for the ssh agent
        Returns the current value of SSH_AUTH_SOCK or None, if the agent could
        not be located/started
        """
        if 'SSH_AUTH_SOCK' not in os.environ:
            proc = subprocess.run(
                'ssh-agent',
                shell=True,
                stdout=subprocess.PIPE
            )
            match = SSH_AGENT_PATTERN.search(proc.stdout.decode())
            if not match:
                return None
            os.environ['SSH_AUTH_SOCK'] = match.group(1)
        return os.environ['SSH_AUTH_SOCK']

    @staticmethod
    def add_key_to_agent(filepath: str):
        """
        Adds the requested filepath to the ssh agent
        This is useful for keys in formats not supported by paramiko/cryptography
        Mojave users: You will probably need to call this function on keys generated by mojave
        """
        if RemoteSlurmBackend.ssh_agent() is None:
            raise RuntimeError("Could not boot ssh agent")
        subprocess.check_call(
            'ssh-add {}'.format(filepath),
            shell=True,
            executable='/bin/bash'
        )

    def __init__(self, hostname: str, **kwargs: typing.Any):
        """
        Initializes the backend.
        No connection is established until the context is entered.
        provided arguments and keyword arguments are passed to paramiko.SSHClient.Connect
        """
        self.hostname = hostname
        self.__hostname = hostname
        self.__sshkwargs = {
            **{
                'timeout': 30,
                'banner_timeout': 30,
                'auth_timeout': 60
            },
            **kwargs
        }
        self.client = paramiko.SSHClient()
        self.client.load_system_host_keys()

    def load_config_args(self):
        """
        Parses the ssh config file and sets up this client's ssh arguments
        as specified by the config file.
        Paramiko is too stupid to read ssh config so we have to write this function
        """
        config_path = os.path.expanduser('~/.ssh/config')
        config = paramiko.SSHConfig()
        if os.path.isfile(config_path):
            with open(config_path) as r:
                config.parse(r)
        config = config.lookup(self.hostname)
        if 'hostname' in config:
            self.hostname = config['hostname']
        if 'port' in config:
            self.__sshkwargs['port'] = config['port']
        if 'user' in config:
            self.__sshkwargs['username'] = config['user']
        if 'identityfile' in config:
            self.__sshkwargs['key_filename'] = os.path.expanduser(config['identityfile'][0])
        if 'userknownhostsfile' in config:
            self.client.load_host_keys(config['userknownhostsfile'])
        if 'hostkeyalias' in config:
            host_keys = self.client.get_host_keys()
            if config['hostkeyalias'] in host_keys:
                host_keys[self.hostname] = host_keys[config['hostkeyalias']]
            else:
                warnings.warn("Requested HostKeyAlias not found. Switching to auto-add policy", stacklevel=2)
                self.client.set_missing_host_key_policy(IgnoreKeyPolicy)
        if 'key_filename' in self.__sshkwargs and ('allow_agent' not in self.__sshkwargs or self.__sshkwargs['allow_agent']):
            try:
                self.add_key_to_agent(self.__sshkwargs['key_filename'])
            except:
                warnings.warn("Unable to add specified key file to ssh agent. Mojave users may be unable to authenticate")

    def _invoke(self, command: str) -> typing.Tuple[paramiko.ChannelFile, paramiko.ChannelFile, paramiko.ChannelFile]:
        """
        Raw handle to exec_command
        """
        if self.client._transport is None:
            raise paramiko.SSHException("Client is not connected")
        return self.client.exec_command(command)

    def invoke(self, command: str, interactive: bool = False) -> typing.Tuple[int, typing.BinaryIO, typing.BinaryIO]:
        """
        Invoke an arbitrary command in the slurm console
        Returns a tuple containing (exit status, byte stream of standard out from the command, byte stream of stderr from the command).
        If interactive is True, stdin, stdout, and stderr should all be connected live to the user's terminal
        """
        raw_stdin, raw_stdout, raw_stderr = self._invoke(command)
        try:
            if interactive:
                return make_interactive(raw_stdout.channel)
            stdout = io.BytesIO(raw_stdout.read())
            stderr = io.BytesIO(raw_stderr.read())
            return raw_stdout.channel.recv_exit_status(), stdout, stderr
        except KeyboardInterrupt:
            print("Warning: Command will continue running on remote server as Paramiko has no way to interrupt commands", file=sys.stderr)
            raise

    def interactive_login(self) -> int:
        """
        Connects to the client interactively.
        Stdin/out/err will be connected directly, and
        Python will not have access to the streams
        Returns ssh's exit status
        """
        return subprocess.call(
            'ssh {}'.format(self.__hostname),
            shell=True,
            executable='/bin/bash'
        )

    def sbcast(self, localpath: str, remotepath: str, *slurmopts: str, **slurmparams: typing.Any):
        """
        Broadcasts the localpath (on the local filesystem)
        to all compute nodes at remotepath
        Additional arguments and keyword arguments provided are passed as slurm arguments to srun
        """
        command = 'sbcast {0} -- {1} {1}'.format(
            ArgumentHelper(*slurmopts, **slurmparams).commandline,
            remotepath
        )
        with self.transport() as transport:
            transport.send(localpath, remotepath)
            status, stdout, stderr = self.invoke(command)
            check_call(command, status, stdout, stderr)

    def __enter__(self):
        """
        Establishes a connection to the remote server
        """
        self.client.connect(self.hostname, **self.__sshkwargs)
        return self

    def __exit__(self, *args):
        """
        Closes the connection to the remote server
        """
        self.client.close()

    def transport(self) -> RemoteTransport:
        """
        Return a Transport object suitable for moving files between the
        SLURM cluster and the local filesystem
        """
        return RemoteTransport(self.client)



from .base import AbstractLocalizer
from .local import BatchedLocalizer, LocalLocalizer
from .remote import RemoteLocalizer

__all__ = [
    'BatchedLocalizer',
    'LocalLocalizer',
    'RemoteLocalizer'
]

# * `CANINE`: The current canine version
# * `CANINE_BACKEND`: The name of the current backend type
# * `CANINE_ADAPTER`: The name of the current adapter type
# * `CANINE_ROOT`: The path to the staging directory
# * `CANINE_COMMON`: The path to the directory where common files are localized
# * `CANINE_OUTPUT`: The path to the directory where job outputs will be staged during delocalization
# * `CANINE_JOB_VARS`: A colon separated list of the names of all variables generated by job inputs
# * `CANINE_JOB_INPUTS`: The path to the directory where job inputs are localized
# * `CANINE_JOB_ROOT`: The path to the working directory for the job. Equal to CWD at the start of the job. Output files should be written here

# $CANINE_ROOT: staging dir
#   /common: common inputs
#   /outputs: output dir (unused?)
#       /{jobID}: job specific output
#   /jobs/{jobID}: job staging dir
#       setup.sh: setup script
#       script.sh: main script
#       teardown.sh: teardown script
#       /inputs: job specific input
#       /workspace: job starting cwd and output dir



import typing
import os
import time
import sys
import warnings
from .adapters import AbstractAdapter, ManualAdapter, FirecloudAdapter
from .backends import AbstractSlurmBackend, LocalSlurmBackend, RemoteSlurmBackend, TransientGCPSlurmBackend
from .localization import AbstractLocalizer, BatchedLocalizer, LocalLocalizer, RemoteLocalizer
import yaml
import pandas as pd
from agutil import status_bar
version = '0.1.1'

ADAPTERS = {
    'Manual': ManualAdapter,
    'Firecloud': FirecloudAdapter,
    'Terra': FirecloudAdapter
}

BACKENDS = {
    'Local': LocalSlurmBackend,
    'Remote': RemoteSlurmBackend,
    'TransientGCP': TransientGCPSlurmBackend
}

LOCALIZERS = {
    'Batched': BatchedLocalizer,
    'Local': LocalLocalizer,
    'Remote': RemoteLocalizer
}

ENTRYPOINT = """#!/bin/bash
export CANINE="{version}"
export CANINE_BACKEND="{{backend}}"
export CANINE_ADAPTER="{{adapter}}"
export CANINE_ROOT="{{CANINE_ROOT}}"
export CANINE_COMMON="{{CANINE_COMMON}}"
export CANINE_OUTPUT="{{CANINE_OUTPUT}}"
export CANINE_JOBS="{{CANINE_JOBS}}"
source $CANINE_JOBS/$SLURM_ARRAY_TASK_ID/setup.sh
{{pipeline_script}}
CANINE_JOB_RC=$?
source $CANINE_JOBS/$SLURM_ARRAY_TASK_ID/teardown.sh
exit $CANINE_JOB_RC
""".format(version=version)

class Orchestrator(object):
    """
    Main class
    Parses a configuration object, initializes, runs, and cleans up a Canine Pipeline
    """

    @staticmethod
    def stringify(obj: typing.Any) -> typing.Any:
        """
        Recurses through the dictionary, converting objects to strings
        """
        if isinstance(obj, list):
            return [
                Orchestrator.stringify(elem)
                for elem in obj
            ]
        elif isinstance(obj, dict):
            return {
                key:Orchestrator.stringify(val)
                for key, val in obj.items()
            }
        return str(obj)

    @staticmethod
    def fill_config(cfg: typing.Union[str, typing.Dict[str, typing.Any]]) -> typing.Dict[str, typing.Any]:
        """
        Loads the given config object (or reads from the given filepath)
        Applies Canine defaults, then returns the final config dictionary
        """
        if isinstance(cfg, str):
            with open(cfg) as r:
                cfg = yaml.load(r, Loader=yaml.loader.SafeLoader)
        DEFAULTS = {
            'name': 'canine',
            'adapter': {
                'type': 'Manual',
            },
            'backend': {
                'type': 'Local'
            },
            'localization': {
                'strategy': 'Batched'
            }
        }
        for key, value in DEFAULTS.items():
            if key not in cfg:
                cfg[key] = value
            elif isinstance(value, dict):
                cfg[key] = {**value, **cfg[key]}
        return cfg


    def __init__(self, config: typing.Union[str, typing.Dict[str, typing.Any]]):
        """
        Initializes the Orchestrator from a given config
        """
        config = Orchestrator.fill_config(config)
        self.name = config['name']
        if 'script' not in config:
            raise KeyError("Config missing required key 'script'")
        self.script = config['script']
        if isinstance(self.script, str):
            if not os.path.isfile(self.script):
                raise FileNotFoundError(self.script)
        elif not isinstance(self.script, list):
            raise TypeError("script must be a path to a bash script or a list of bash commands")
        self.raw_inputs = Orchestrator.stringify(config['inputs']) if 'inputs' in config else {}
        self.resources = Orchestrator.stringify(config['resources']) if 'resources' in config else {}
        adapter = config['adapter']
        if adapter['type'] not in ADAPTERS:
            raise ValueError("Unknown adapter type '{}'".format(adapter))
        self._adapter_type=adapter['type']
        self.adapter = ADAPTERS[adapter['type']](**{arg:val for arg,val in adapter.items() if arg != 'type'})
        backend = config['backend']
        if backend['type'] not in BACKENDS:
            raise ValueError("Unknown backend type '{}'".format(backend))
        self._backend_type=backend['type']
        self.backend = BACKENDS[backend['type']](**{arg:val for arg,val in backend.items() if arg != 'type'})
        self.localizer_args = config['localization'] if 'localization' in config else {}
        if self.localizer_args['strategy'] not in LOCALIZERS:
            raise ValueError("Unknown localization strategy '{}'".format(self.localizer_args))
        self._localizer_type = LOCALIZERS[self.localizer_args['strategy']]
        self.localizer_args = {key:val for key,val in self.localizer_args.items() if key != 'strategy'}
        self.localizer_overrides = {}
        if 'overrides' in self.localizer_args:
            self.localizer_overrides = {**self.localizer_args['overrides']}
            del self.localizer_args['overrides']
        self.raw_outputs = Orchestrator.stringify(config['outputs']) if 'outputs' in config else {}
        if len(self.raw_outputs) == 0:
            warnings.warn("No outputs declared", stacklevel=2)
        if 'stdout' not in self.raw_outputs:
            self.raw_outputs['stdout'] = 'stdout'
        if 'stderr' not in self.raw_outputs:
            self.raw_outputs['stderr'] = 'stderr'

    def run_pipeline(self, output_dir: str = 'canine_output', dry_run: bool = False) -> typing.Tuple[str, dict, dict, pd.DataFrame]:
        """
        Runs the configured pipeline
        Returns a 4-tuple:
        * The batch job id
        * The input job specification
        * The sacct dataframe after all jobs completed
        """
        job_spec = self.adapter.parse_inputs(self.raw_inputs)
        print("Preparing pipeline of", len(job_spec), "jobs")
        print("Connecting to backend...")
        if isinstance(self.backend, RemoteSlurmBackend):
            self.backend.load_config_args()
        with self.backend:
            print("Initializing pipeline workspace")
            with self._localizer_type(self.backend, **self.localizer_args) as localizer:
                print("Localizing inputs...")
                abs_staging_dir = localizer.localize(
                    job_spec,
                    self.raw_outputs,
                    self.localizer_overrides
                )
                print("Job staged on SLURM controller in:", abs_staging_dir)
                print("Preparing pipeline script")
                env = localizer.environment('compute')
                root_dir = env['CANINE_ROOT']
                entrypoint_path = os.path.join(root_dir, 'entrypoint.sh')
                if isinstance(self.script, str):
                    pipeline_path = os.path.join(root_dir, os.path.basename(self.script))
                else:
                    pipeline_path = self.backend.pack_batch_script(
                        *self.script,
                        script_path=os.path.join(root_dir, 'script.sh')
                    )
                with self.backend.transport() as transport:
                    if isinstance(self.script, str):
                        transport.send(self.script, pipeline_path)
                        transport.chmod(pipeline_path, 0o775)
                    with transport.open(entrypoint_path, 'w') as w:
                        w.write(ENTRYPOINT.format(
                            backend=self._backend_type,
                            adapter=self._adapter_type,
                            pipeline_script=pipeline_path,
                            **env
                        ))
                    transport.chmod(entrypoint_path, 0o775)
                    transport.chmod(pipeline_path, 0o775)
                if dry_run:
                    localizer.clean_on_exit = False
                    return job_spec
                print("Waiting for cluster to finish startup...")
                self.backend.wait_for_cluster_ready()
                print("Submitting batch job")
                batch_id = self.backend.sbatch(
                    entrypoint_path,
                    'requeue',
                    job_name=self.name,
                    array="0-{}".format(len(job_spec)-1),
                    output="{}/%a/workspace/stdout".format(env['CANINE_JOBS']),
                    error="{}/%a/workspace/stderr".format(env['CANINE_JOBS']),
                    **self.resources
                )
                print("Batch id:", batch_id)
                completed_jobs = []
                try:
                    waiting_jobs = {
                        '{}_{}'.format(batch_id, i)
                        for i in range(len(job_spec))
                    }
                    outputs = {}
                    while len(waiting_jobs):
                        time.sleep(30)
                        acct = self.backend.sacct()
                        for jid in [*waiting_jobs]:
                            if jid in acct.index and acct['State'][jid] not in {'RUNNING', 'PENDING', 'NODE_FAIL'}:
                                job = jid.split('_')[1]
                                print("Job",job, "completed with status", acct['State'][jid], acct['ExitCode'][jid].split(':')[0])
                                # outputs.update(localizer.delocalize(self.raw_outputs, jobId=job))
                                # waiting_jobs.remove(jid)
                                completed_jobs.append((job, jid))
                                waiting_jobs.remove(jid)
                except:
                    print("Encountered unhandled exception. Cancelling batch job", file=sys.stderr)
                    self.backend.scancel(batch_id)
                    raise
                finally:
                    if len(completed_jobs):
                        print("Delocalizing outputs")
                        outputs = localizer.delocalize(self.raw_outputs, output_dir)
            print("Parsing output data")
            self.adapter.parse_outputs(outputs)
            return batch_id, job_spec, outputs, self.backend.sacct()



import typing
import os
import sys
import select
import io
import warnings
import shlex
from subprocess import CalledProcessError
import google.auth
import paramiko

class ArgumentHelper(dict):
    """
    Helper class for setting arguments to slurm commands
    Used only to handle keyword arguments to console commands
    Should not be responsible for positionals
    """

    def __init__(self, *flags: str, **params: typing.Any):
        """
        Creates a new ArgumentHelper
        Flags can be passed as positional arguments
        Parameters can be passed as keyword arguments
        """
        object.__setattr__(self, 'defaults', {})
        object.__setattr__(self, 'flags', [item for item in flags])
        object.__setattr__(self, 'params', {k:v for k,v in params.items()})
        for key, val in [*self.params.items()]:
            if val is True:
                self.flags.append(key)
                del self.params[key]

    def __repr__(self) -> str:
        return '<ArgumentHelper{}>'.format(
            self.commandline
        )

    def __getattr__(self, name):
        return self[name]

    def __setattr__(self, name, value):
        self[name] = value

    def __delattr__(self, name):
        del self[name]

    def __getitem__(self, name):
        if name in self.params:
            return self.params[name]
        if name in self.flags:
            return True
        if name in self.defaults:
            return self.defaults[name]

    def __setitem__(self, name, value):
        if value is True:
            self.flags.append(name)
        elif value is False:
            object.__setattr__(self, 'flags', [flag for flag in self.flags if flag != name])
        else:
            self.params[name] = value

    def __delitem__(self, name):
        if name in self.params:
            del self.params[name]
        elif name in self.flags:
            object.__setattr__(self, 'flags', [flag for flag in self.flags if flag != name])
        else:
            raise KeyError("No such argument {}".format(name))

    @staticmethod
    def translate(flag) -> str:
        """Converts acceptable python strings to command line args"""
        return flag.replace('_', '-')

    @property
    def commandline(self) -> str:
        """Expands the arguments to command line form"""
        return '{short_prespace}{short_flags}{long_flags}{params}'.format(
            short_prespace=' -' if len([f for f in self.flags if len(f) == 1]) else '',
            short_flags=''.join(flag for flag in self.flags if len(flag)==1),
            long_flags=''.join(' --{}'.format(self.translate(flag)) for flag in self.flags if len(flag) > 1),
            params=''.join(
                ' --{}={}'.format(self.translate(key), shlex.quote(value))
                for key, value in self.params.items()
            )
        )

    def setdefaults(self, **kwargs: typing.Any):
        self.defaults.update(kwargs)

def make_interactive(channel: paramiko.Channel) -> typing.Tuple[int, typing.BinaryIO, typing.BinaryIO]:
    """
    Manages an interactive command
    Takes in a paramiko.Channel shared by stdin, stdout, stderr of a currently running command
    The current interpreter stdin is duplicated and written to the command's stdin
    The command's stdout and stderr are written to the interpreter's stdout and stderr
    and also buffered in a ByteStream for later reading
    Returns (exit status, Stdout buffer, Stderr bufer)
    """
    infd = os.dup(sys.stdin.fileno())
    try:
        poll = select.poll()
        poll.register(infd)
        poll.register(channel, select.POLLIN+select.POLLPRI+select.POLLERR+select.POLLHUP)
        stdout = io.BytesIO()
        stderr = io.BytesIO()
        while not channel.exit_status_ready():
            for fd, event in poll.poll(0.5):
                if fd == infd and event == select.POLLIN or event == select.POLLPRI:
                    # Text available on python stdin
                    channel.send(os.read(infd, 4096))
            if channel.recv_ready():
                content = channel.recv(4096)
                sys.stdout.write(content.decode())
                stdout.write(content)
            if channel.recv_stderr_ready():
                content = channel.recv_stderr(4096)
                sys.stderr.write(content.decode())
                stderr.write(content)
        if channel.recv_ready():
            content = channel.recv(4096)
            sys.stdout.write(content.decode())
            stdout.write(content)
        if channel.recv_stderr_ready():
            content = channel.recv_stderr(4096)
            sys.stderr.write(content.decode())
            stderr.write(content)
        stdout.seek(0,0)
        stderr.seek(0,0)
        return channel.recv_exit_status(), stdout, stderr
    finally:
        os.close(infd)

__DEFAULT_GCP_PROJECT__ = None

def get_default_gcp_project():
    """
    Returns the currently configured default project
    """
    global __DEFAULT_GCP_PROJECT__
    if __DEFAULT_GCP_PROJECT__ is None:
        __DEFAULT_GCP_PROJECT__ = google.auth.default()[1]
    return __DEFAULT_GCP_PROJECT__

def check_call(cmd:str, rc: int, stdout: typing.Optional[typing.BinaryIO] = None, stderr: typing.Optional[typing.BinaryIO] = None):
    """
    Checks that the rc is 0
    If not, flush stdout and stderr streams and raise a CalledProcessError
    """
    if rc != 0:
        if stdout is not None:
            sys.stdout.write(stdout.read().decode())
            sys.stdout.flush()
        if stderr is not None:
            sys.stderr.write(stderr.read().decode())
            sys.stderr.flush()
        raise CalledProcessError(rc, cmd)



from __future__ import print_function
import argparse
import glob
import os
import shutil

"""
This is not actually part of the canine package
This is helper code which is run in the backend
"""

def main(dest, jobId, patterns):
    jobdir = os.path.join(dest, str(jobId))
    if not os.path.isdir(jobdir):
        os.makedirs(jobdir)
    for name, pattern in patterns:
        for target in glob.iglob(pattern):
            dest = os.path.join(jobdir, name, os.path.basename(target))
            if not os.path.exists(dest):
                if not os.path.isdir(os.path.dirname(dest)):
                    os.makedirs(os.path.dirname(dest))
                if os.path.isfile(target):
                    os.symlink(os.path.abspath(target), dest)
                else:
                    shutil.copytree(target, dest)

if __name__ == '__main__':
    parser = argparse.ArgumentParser('canine-delocalizer')
    parser.add_argument(
        'dest',
        help="Destination directory",
    )
    parser.add_argument(
        'jobId',
        help="Job ID"
    )
    parser.add_argument(
        '-p', '--pattern',
        nargs=2,
        action='append',
        help="Pattern name and pattern. in form '-p {name} {pattern}'",
        default=[]
    )
    args = parser.parse_args()
    main(args.dest, args.jobId, args.pattern)



import typing
import json
import os
import warnings
import pandas as pd
import numpy as np
from .base import AbstractAdapter
import dalmatian

class FirecloudAdapter(AbstractAdapter):
    """
    Job input adapter
    Parses inputs as firecloud expression bois
    if enabled, job outputs will be written back to workspace
    """
    def __init__(self, workspace: str, entityType: str, entityName: str, entityExpression: typing.Optional[str] = None, write_to_workspace: bool = True):
        """
        Initializes the adapter
        Must provide workspace and entity information
        If no expression is provided:
        * Assume a single job, and resolve all input expressions in the context of the one entity
        If an expression is provided:
        * Assume multiple entities (entity type will be auto-detected)
        * Launch one job per entity, resolving input expressions for each one
        """
        self.workspace = dalmatian.WorkspaceManager(workspace)
        self.workspace.populate_cache()
        self.workspace.go_offline()
        if entityName not in self.workspace._get_entities_internal(entityType).index:
            raise NameError('No such {} "{}" in workspace {}'.format(
                entityType,
                entityName,
                workspace
            ))
        self._entityType = entityType
        self._entityName = entityName
        self._entityExpression = entityExpression
        self.evaluator = dalmatian.Evaluator(self.workspace.entity_types)
        for etype in self.workspace.entity_types:
            self.evaluator.add_entities(etype, self.workspace._get_entities_internal(etype))
        self.evaluator.add_attributes(self.workspace.attributes)
        if entityExpression is not None:
            self.entities = self.evaluator(entityType, entityName, entityExpression)
            self.etype = self.evaluator.determine_reference_type(entityType, self.entities, '')
        else:
            self.entities = [entityName]
            self.etype = entityType
        self.write_to_workspace = write_to_workspace
        self.__spec = None
        print("initialized fc adapter in workspace", self.workspace)
        print(len(self.entities), "of type", self.etype)

    def evaluate(self, etype: str, entity: str, expr: str) -> str:
        """
        Evaluates and unpacks the results of a firecloud expression to a single value
        Raises appropriate exceptions otherwise
        """
        results = [
            item for item in self.evaluator(etype, entity, expr)
            if isinstance(item, str) or not np.isnan(item)
        ]

        if len(results) == 1:
            return results[0]
        elif len(results) == 0:
            return None
        else:
            raise ValueError("Expression '{}' on {} {} returned more than one result".format(
                expr,
                etype,
                entity
            ))

    def parse_inputs(self, inputs: typing.Dict[str, typing.Union[typing.Any, typing.List[typing.Any]]]) -> typing.Dict[str, typing.Dict[str, str]]:
        """
        Takes raw user inputs and parses out actual inputs for each job
        Returns a job input specification useable for Localization
        Also sets self.spec to the same dictionary
        """
        # If constant input:
        #   this. or workspace. -> evaluate
        #   gs:// -> raw
        #   else: warn, but assume raw
        # if array input: fail (can't map array expr to entity expr)
        self.__spec = {str(i):{} for i in range(len(self.entities))}
        for name, expr in inputs.items():
            if isinstance(expr, list):
                raise TypeError("Firecloud adapter cannot handle array-type inputs")
            elif expr.startswith('this.') or expr.startswith('workspace.'):
                for i, entity in enumerate(self.entities):
                    self.__spec[str(i)][name] = self.evaluate(self.etype, entity, expr)
            else:
                if isinstance(expr, str) and not expr.startswith('gs://'):
                    warnings.warn("Assuming expression is not a Firecloud expression", stacklevel=2)
                for i, entity in enumerate(self.entities):
                    self.__spec[str(i)][name] = expr
        self.workspace.hound.write_log_entry(
            'job',
            'Canine launching new job with input configuration: {}; Results will{} be written back to workspace'.format(
                json.dumps({
                    entity:cfg for entity, (i, cfg) in zip(self.entities, self.__spec.items())
                }),
                '' if self.write_to_workspace else ' not'
            ),
            entities=[
                '{}/{}'.format(self.etype, entity)
                for entity in self.entities
            ]
        )
        return self.spec

    def parse_outputs(self, outputs: typing.Dict[str, typing.Dict[str, typing.List[str]]]):
        """
        Takes a dictionary of job outputs
        {jobId: {outputName: [output paths]}}
        If write_to_workspace is enabled, output files will be written back to firecloud
        """
        if self.write_to_workspace:
            with self.workspace.hound.with_reason('Uploading results from Canine job'):
                outputDf = pd.DataFrame.from_dict(
                    {
                        jobId: {
                            outputName: [outputFile for outputFile in outputFiles if os.path.exists(outputFile)]
                            for outputName, outputFiles in jobOutput.items()
                            if outputName not in {'stdout', 'stderr'}
                        }
                        for jobId, jobOutput in outputs.items()
                    },
                    orient='index'
                ).set_index(
                    pd.Index(
                        [self.entities[int(i)] for i in outputs],
                        name='{}_id'.format(self.etype)
                    )
                ).applymap(
                    lambda cell: cell if not isinstance(cell, list) else (
                        cell[0] if len(cell) > 0 else np.nan
                    )
                )
                self.workspace.sync()
                self.workspace.update_entity_attributes(
                    self.etype,
                    outputDf
                )

    @property
    def spec(self) -> typing.Dict[str, typing.Dict[str, str]]:
        """
        The most recent job specification
        """
        return {
            jobId: {**spec}
            for jobId, spec in self.__spec.items()
        }



"""
canine.backends
=================================
Contains SLURM backends for canine
"""

from .base import AbstractTransport, AbstractSlurmBackend
from .local import LocalTransport, LocalSlurmBackend
from .remote import RemoteTransport, RemoteSlurmBackend
from .gcpTransient import TransientGCPSlurmBackend

__all__ = [
    'LocalSlurmBackend',
    'LocalTransport',
    'RemoteSlurmBackend',
    'RemoteTransport',
    'TransientGCPSlurmBackend'
]



import abc
import os
import sys
import warnings
import typing
import fnmatch
import shlex
import tempfile
import subprocess
import traceback
import shutil
from uuid import uuid4
from collections import namedtuple
from contextlib import ExitStack, contextmanager
from ..backends import AbstractSlurmBackend, AbstractTransport, LocalSlurmBackend
from ..utils import get_default_gcp_project, check_call
from hound.client import _getblob_bucket
from agutil import status_bar

Localization = namedtuple("Localization", ['type', 'path'])
# types: stream, download, None
# indicates what kind of action needs to be taken during job startup

PathType = namedtuple(
    'PathType',
    ['localpath', 'controllerpath', 'computepath']
)

class AbstractLocalizer(abc.ABC):
    """
    Base class for localization.
    """
    requester_pays = {}

    def __init__(self, backend: AbstractSlurmBackend, strategy: str = "batched", transfer_bucket: typing.Optional[str] = None, common: bool = True, staging_dir: str = None, mount_path: str = None, localize_gs: bool = None):
        """
        Initializes the Localizer using the given transport.
        Localizer assumes that the SLURMBackend is connected and functional during
        the localizer's entire life cycle.
        If staging_dir is not provided, a random directory is chosen
        """
        if localize_gs is not None:
            warnings.warn("localize_gs option removed and ignored. Use overrides to explicitly control this behavior")
        self.transfer_bucket = transfer_bucket
        if transfer_bucket is not None and self.transfer_bucket.startswith('gs://'):
            self.transfer_bucket = self.transfer_bucket[5:]
        self.strategy = strategy.lower()
        self.backend = backend
        self.common = common
        self.common_inputs = set()
        self.__sbcast = False
        if staging_dir == 'SBCAST':
            # FIXME: This doesn't actually do anything yet
            # If sbcast is true, then localization needs to use backend.sbcast to move files to the remote system
            # Not sure at all how delocalization would work
            self.__sbcast = True
            staging_dir = None
        self.staging_dir = staging_dir if staging_dir is not None else str(uuid4())
        self._local_dir = None if self.strategy == 'remote' else tempfile.TemporaryDirectory()
        self.local_dir = None if self.strategy == 'remote' else self._local_dir.name
        with self.backend.transport() as transport:
            self.mount_path = transport.normpath(mount_path if mount_path is not None else self.staging_dir)
        self.inputs = {} # {jobId: {inputName: (handle type, handle value)}}
        self.clean_on_exit = True

    def get_requester_pays(self, path: str) -> bool:
        """
        Returns True if the requested gs:// object or bucket resides in a
        requester pays bucket
        """
        if path.startswith('gs://'):
            path = path[5:]
        bucket = path.split('/')[0]
        if bucket not in self.requester_pays:
            command = 'gsutil ls gs://{}'.format(path)
            try:
                # We check on the remote host because scope differences may cause
                # a requester pays bucket owned by this account to require -u on the controller
                # better safe than sorry
                rc, sout, serr = self.backend.invoke(command)
                self.requester_pays[bucket] = len([line for line in serr.readlines() if b'requester pays bucket but no user project provided' in line]) >= 1
            except subprocess.CalledProcessError:
                pass
        return bucket in self.requester_pays and self.requester_pays[bucket]

    @contextmanager
    def transport_context(self, transport: typing.Optional[AbstractTransport] = None) -> typing.ContextManager[AbstractTransport]:
        """
        Opens a file transport in the context.
        If an existing transport is provided, it is passed through
        If None (default) is provided, a new transport is opened, and closed after the context
        """
        with ExitStack() as stack:
            if transport is None:
                transport = stack.enter_context(self.backend.transport())
            yield transport

    def environment(self, location: str) -> typing.Dict[str, str]:
        """
        Returns environment variables relative to the given location.
        Location must be one of {"local", "controller", "compute"}
        """
        if location not in {"local", "controller", "compute"}:
            raise ValueError('location must be one of {"local", "controller", "compute"}')
        if location == 'local':
            return {
                'CANINE_ROOT': self.local_dir,
                'CANINE_COMMON': os.path.join(self.local_dir, 'common'),
                'CANINE_OUTPUT': os.path.join(self.local_dir, 'outputs'), #outputs/jobid/outputname/...files...
                'CANINE_JOBS': os.path.join(self.local_dir, 'jobs'),
            }
        elif location == "controller":
            return {
                'CANINE_ROOT': self.staging_dir,
                'CANINE_COMMON': os.path.join(self.staging_dir, 'common'),
                'CANINE_OUTPUT': os.path.join(self.staging_dir, 'outputs'), #outputs/jobid/outputname/...files...
                'CANINE_JOBS': os.path.join(self.staging_dir, 'jobs'),
            }
        elif location == "compute":
            return {
                'CANINE_ROOT': self.mount_path,
                'CANINE_COMMON': os.path.join(self.mount_path, 'common'),
                'CANINE_OUTPUT': os.path.join(self.mount_path, 'outputs'),
                'CANINE_JOBS': os.path.join(self.mount_path, 'jobs'),
            }
    def gs_dircp(self, src: str, dest: str, context: str, transport: typing.Optional[AbstractTransport] = None):
        """
        gs_copy for directories
        context must be one of {'local', 'remote'}, which specifies
        where the command should be run
        When uploading to gs://, the destination gs:// director does not have to exist
        When downloading from gs:// the destination directory does not have to exist,
        but its parent does
        """
        assert context in {'local', 'remote'}
        gs_obj = src if src.startswith('gs://') else dest
        if not dest.startswith('gs://'):
            # Download
            if context == 'remote':
                with self.transport_context(transport) as transport:
                    if not transport.exists(dest):
                        transport.makedirs(dest)
            else:
                if not os.path.exists(dest):
                    os.makedirs(dest)
        if not src.startswith('gs://'):
            # Upload
            # Fix empty dirs by touching a file
            # Won't fix nested directory structure containing empty dirs,
            # but it seems inefficient to walk and touch directories
            if context == 'remote':
                self.backend.invoke('touch {}/.canine_dir_marker'.format(src))
            else:
                subprocess.run(['touch', '{}/.canine_dir_marker'.format(src)])
        command = "gsutil -m -o GSUtil:check_hashes=if_fast_else_skip -o GSUtil:parallel_composite_upload_threshold=150M {} cp -r {} {}".format(
            '-u {}'.format(get_default_gcp_project()) if self.get_requester_pays(gs_obj) else '',
            src,
            dest
        )
        if context == 'remote':
            # Invoke interactively
            rc, sout, serr = self.backend.invoke(command, True)
            check_call(command, rc, sout, serr)
        else:
            subprocess.check_call(command, shell=True)
        if not dest.startswith('gs://'):
            # Clean up .dir file
            if context == 'remote':
                self.backend.invoke('rm -f {}/*/.canine_dir_marker'.format(dest))
            else:
                subprocess.run(['rm', '-f', '{}/*/.canine_dir_marker'.format(dest)])

    def gs_copy(self, src: str, dest: str, context: str):
        """
        Copy a google storage (gs://) object
        context must be one of {'local', 'remote'}, which specifies
        where the command should be run
        When uploading to gs://, the destination gs:// directory does not have to exist
        When downloading from gs:// the destination parent directory must exist
        """
        assert context in {'local', 'remote'}
        gs_obj = src if src.startswith('gs://') else dest
        try:
            components = gs_obj[5:].split('/')
            bucket = _getblob_bucket(None, components[0], None)
            blobs = {
                blob.name
                for page in bucket.list_blobs(prefix='/'.join(components[1:]), fields='items/name,nextPageToken').pages
                for blob in page
            }
            if len([blob for blob in blobs if blob == '/'.join(components[1:])]) == 0:
                # This is a directory
                print("Copying directory:", gs_obj)
                return self.gs_dircp(src, os.path.dirname(dest), context)
        except:
            # If there is an exception, or the above condition is false
            # Procede as a regular gs_copy
            traceback.print_exc()

        command = "gsutil -o GSUtil:check_hashes=if_fast_else_skip -o GSUtil:parallel_composite_upload_threshold=150M {} cp {} {}".format(
            '-u {}'.format(get_default_gcp_project()) if self.get_requester_pays(gs_obj) else '',
            src,
            dest
        )
        if context == 'remote':
            rc, sout, serr = self.backend.invoke(command, True)
            check_call(command, rc, sout, serr)
        else:
            subprocess.check_call(command, shell=True)

    def sendtree(self, src: str, dest: str, transport: typing.Optional[AbstractTransport] = None, exist_okay=False):
        """
        Transfers the given local folder to the given remote destination.
        Source must be a local folder, and destination must not exist
        """
        if isinstance(self.backend, LocalSlurmBackend):
            if exist_okay:
                print("exist_okay not supported by LocalSlurmBackend", file=sys.stderr)
            return shutil.copytree(src, dest)
        with self.transport_context(transport) as transport:
            if not os.path.isdir(src):
                raise ValueError("Not a directory: "+src)
            if transport.exists(dest) and not exist_okay:
                raise ValueError("Destination already exists: "+dest)
            dest = transport.normpath(dest)
            if self.transfer_bucket is not None:
                print("Transferring through bucket", self.transfer_bucket)
                path = os.path.join(str(uuid4()), os.path.basename(dest))
                self.gs_dircp(
                    src,
                    'gs://{}/{}'.format(self.transfer_bucket, path),
                    'local',
                    transport=transport
                )
                if not transport.isdir(os.path.dirname(dest)):
                    transport.makedirs(os.path.dirname(dest))
                self.gs_dircp(
                    'gs://{}/{}'.format(self.transfer_bucket, path),
                    os.path.dirname(dest),
                    'remote',
                    transport=transport
                )
                cmd = "gsutil -m {} rm -r gs://{}/{}".format(
                    '-u {}'.format(get_default_gcp_project()) if self.get_requester_pays(self.transfer_bucket) else '',
                    self.transfer_bucket,
                    os.path.dirname(path),
                )
                print(cmd)
                subprocess.check_call(
                    cmd,
                    shell=True
                )
            else:
                print("Transferring directly over SFTP")
                transport.sendtree(src, dest)

    def receivetree(self, src: str, dest: str, transport: typing.Optional[AbstractTransport] = None, exist_okay=False):
        """
        Transfers the given remote folder to the given local destination.
        Source must be a remote folder, and dest must not exist
        """
        if isinstance(self.backend, LocalSlurmBackend):
            if exist_okay:
                print("exist_okay not supported by LocalSlurmBackend", file=sys.stderr)
            return shutil.copytree(src, dest)
        with self.transport_context(transport) as transport:
            if not transport.isdir(src):
                raise ValueError("Not a directory: "+src)
            if os.path.exists(dest) and not exist_okay:
                raise ValueError("Destination already exists: "+dest)
            dest = os.path.abspath(dest)
            if self.transfer_bucket is not None:
                print("Transferring through bucket", self.transfer_bucket)
                path = os.path.join(str(uuid4()), os.path.basename(dest))
                self.gs_dircp(
                    src,
                    'gs://{}/{}'.format(self.transfer_bucket, path),
                    'remote',
                    transport=transport
                )
                if not os.path.isdir(os.path.dirname(dest)):
                    os.makedirs(os.path.dirname(dest))
                self.gs_dircp(
                    'gs://{}/{}'.format(self.transfer_bucket, path),
                    os.path.dirname(dest),
                    'local',
                    transport=transport
                )
                cmd = "gsutil -m {} rm -r gs://{}/{}".format(
                    '-u {}'.format(get_default_gcp_project()) if self.get_requester_pays(self.transfer_bucket) else '',
                    self.transfer_bucket,
                    os.path.dirname(path),
                )
                print(cmd)
                subprocess.check_call(
                    cmd,
                    shell=True
                )
            else:
                print("Transferring directly over SFTP")
                transport.receivetree(src, dest)

    def reserve_path(self, *args: typing.Any) -> PathType:
        """
        Takes any number of path components, relative to the CANINE_ROOT directory
        Returns a PathType object which can be used to view the path relative to
        local, compute, or controller contexts
        """
        return PathType(
            os.path.join(self.environment('local')['CANINE_ROOT'], *(str(component) for component in args)),
            os.path.join(self.environment('controller')['CANINE_ROOT'], *(str(component) for component in args)),
            os.path.join(self.environment('compute')['CANINE_ROOT'], *(str(component) for component in args))
        )

    def delocalize(self, patterns: typing.Dict[str, str], output_dir: str = 'canine_output') -> typing.Dict[str, typing.Dict[str, str]]:
        """
        Delocalizes output from all jobs
        """
        self.receivetree(
            self.environment('controller')['CANINE_OUTPUT'],
            output_dir,
            exist_okay=True
        )
        output_files = {}
        for jobId in os.listdir(output_dir):
            start_dir = os.path.join(output_dir, jobId)
            if not os.path.isdir(start_dir):
                continue
            output_files[jobId] = {}
            for dirpath, dirnames, filenames in os.walk(start_dir):
                for name, pattern in patterns.items():
                    for filename in filenames:
                        fullname = os.path.join(dirpath, filename)
                        if fnmatch.fnmatch(fullname, pattern) or fnmatch.fnmatch(os.path.relpath(fullname, start_dir), pattern):
                            if name not in output_files[jobId]:
                                output_files[jobId][name] = [os.path.abspath(fullname)]
                            else:
                                output_files[jobId][name].append(os.path.abspath(fullname))
        return output_files

    def pick_common_inputs(self, inputs: typing.Dict[str, typing.Dict[str, str]], overrides: typing.Dict[str, typing.Optional[str]], transport: typing.Optional[AbstractTransport] = None) -> typing.Dict[str, str]:
        """
        Scans input configuration and overrides to choose inputs which should be treated as common.
        Returns the dictionary of common inputs {input path: common path}
        """
        with self.transport_context(transport) as transport:
            self.common_inputs = set()
            seen = set()
            for jobId, values in inputs.items():
                for arg, path in values.items():
                    if path in seen and (arg not in overrides or overrides[arg] == 'common'):
                        self.common_inputs.add(path)
                    if arg in overrides and overrides[arg] == 'common':
                        self.common_inputs.add(path)
                    seen.add(path)
            common_dests = {}
            for path in self.common_inputs:
                if path.startswith('gs://') or os.path.exists(path):
                    common_dests[path] = self.reserve_path('common', os.path.basename(os.path.abspath(path)))
                    self.localize_file(path, common_dests[path], transport=transport)
                else:
                    print("Could not handle common file", path, file=sys.stderr)
            return {key: value for key, value in common_dests.items()}

    def finalize_staging_dir(self, jobs: typing.Iterable[str], transport: typing.Optional[AbstractTransport] = None) -> str:
        """
        Finalizes the staging directory by building any missing directory trees
        (such as those dropped during transfer for being empty).
        Returns the absolute path of the remote staging directory on the controller node.
        """
        controller_env = self.environment('controller')
        with self.transport_context(transport) as transport:
            if not transport.isdir(controller_env['CANINE_COMMON']):
                transport.mkdir(controller_env['CANINE_COMMON'])
            if not transport.isdir(controller_env['CANINE_JOBS']):
                transport.mkdir(controller_env['CANINE_JOBS'])
            print("Finalizing directory structure. This may take a while...")
            for jobId in status_bar.iter(jobs):
                if not transport.isdir(os.path.join(controller_env['CANINE_JOBS'], jobId, 'workspace')):
                    transport.makedirs(os.path.join(controller_env['CANINE_JOBS'], jobId, 'workspace'))
                if not transport.isdir(os.path.join(controller_env['CANINE_JOBS'], jobId, 'inputs')):
                    transport.makedirs(os.path.join(controller_env['CANINE_JOBS'], jobId, 'inputs'))
            if not transport.isdir(controller_env['CANINE_OUTPUT']):
                transport.mkdir(controller_env['CANINE_OUTPUT'])
            return transport.normpath(self.staging_dir)

    def prepare_job_inputs(self, jobId: str, job_inputs: typing.Dict[str, str], common_dests: typing.Dict[str, str], overrides: typing.Dict[str, typing.Optional[str]], transport: typing.Optional[AbstractTransport] = None):
        """
        Prepares job-specific inputs.
        Fills self.inputs[jobId] with Localization objects for each input
        """
        with self.transport_context(transport) as transport:
            self.inputs[jobId] = {}
            for arg, value in job_inputs.items():
                mode = overrides[arg] if arg in overrides else False
                if mode is not False:
                    if mode == 'stream':
                        if not value.startswith('gs://'):
                            print("Ignoring 'stream' override for", arg, "with value", value, "and localizing now", file=sys.stderr)
                            self.inputs[jobId][arg] = Localization(
                                None,
                                self.reserve_path('jobs', jobId, 'inputs', os.path.basename(os.path.abspath(value)))
                            )
                            self.localize_file(
                                value,
                                self.inputs[jobId][arg].path,
                                transport=transport
                            )
                        else:
                            self.inputs[jobId][arg] = Localization(
                                'stream',
                                value
                            )
                    elif mode == 'localize':
                        self.inputs[jobId][arg] = Localization(
                            None,
                            self.reserve_path('jobs', jobId, 'inputs', os.path.basename(os.path.abspath(value)))
                        )
                        self.localize_file(
                            value,
                            self.inputs[jobId][arg].path,
                            transport=transport
                        )
                    elif mode == 'delayed':
                        if not value.startswith('gs://'):
                            print("Ignoring 'delayed' override for", arg, "with value", value, "and localizing now", file=sys.stderr)
                            self.inputs[jobId][arg] = Localization(
                                None,
                                self.reserve_path('jobs', jobId, 'inputs', os.path.basename(os.path.abspath(value)))
                            )
                            self.localize_file(
                                value,
                                self.inputs[jobId][arg].path,
                                transport=transport
                            )
                        else:
                            self.inputs[jobId][arg] = Localization(
                                'download',
                                value
                            )
                    elif mode is None:
                        # Do not reserve path here
                        # null override treats input as string
                        self.inputs[jobId][arg] = Localization(None, value)
                elif value in common_dests:
                    # common override already handled
                    # No localization needed, already copied
                    self.inputs[jobId][arg] = Localization(None, common_dests[value])
                else:
                    if os.path.exists(value) or value.startswith('gs://'):
                        remote_path = self.reserve_path('jobs', jobId, 'inputs', os.path.basename(os.path.abspath(value)))
                        self.localize_file(value, remote_path, transport=transport)
                        value = remote_path
                    self.inputs[jobId][arg] = Localization(
                        None,
                        # value will either be a PathType, if localized above,
                        # or an unchanged string if not handled
                        value
                    )

    def job_setup_teardown(self, jobId: str, patterns: typing.Dict[str, str]) -> typing.Tuple[str, str]:
        """
        Returns a tuple of (setup script, teardown script) for the given job id.
        Must call after pre-scanning inputs
        """
        job_vars = []
        exports = []
        extra_tasks = [
            'if [[ -d $CANINE_JOB_INPUTS ]]; then cd $CANINE_JOB_INPUTS; fi'
        ]
        compute_env = self.environment('compute')
        for key, val in self.inputs[jobId].items():
            if val.type == 'stream':
                job_vars.append(shlex.quote(key))
                dest = self.reserve_path('jobs', jobId, 'inputs', os.path.basename(os.path.abspath(val.path)))
                extra_tasks += [
                    'if [[ -e {0} ]]; then rm {0}; fi'.format(dest.computepath),
                    'mkfifo {}'.format(dest.computepath),
                    "gsutil {} cat {} > {} &".format(
                        '-u {}'.format(shlex.quote(get_default_gcp_project())) if self.get_requester_pays(val.path) else '',
                        shlex.quote(val.path),
                        dest.computepath
                    )
                ]
                exports.append('export {}="{}"'.format(
                    key,
                    dest.computepath
                ))
            elif val.type == 'download':
                job_vars.append(shlex.quote(key))
                dest = self.reserve_path('jobs', jobId, 'inputs', os.path.basename(os.path.abspath(val.path)))
                extra_tasks += [
                    "if [[ ! -e {2}.fin ]]; then gsutil {0} -o GSUtil:check_hashes=if_fast_else_skip cp {1} {2} && touch {2}.fin; fi".format(
                        '-u {}'.format(shlex.quote(get_default_gcp_project())) if self.get_requester_pays(val.path) else '',
                        shlex.quote(val.path),
                        dest.computepath
                    )
                ]
                exports.append('export {}="{}"'.format(
                    key,
                    dest.computepath
                ))
            elif val.type is None:
                job_vars.append(shlex.quote(key))
                exports.append('export {}="{}"'.format(
                    key,
                    shlex.quote(val.path.computepath if isinstance(val.path, PathType) else val.path)
                ))
            else:
                print("Unknown localization command:", val.type, "skipping", key, val.path, file=sys.stderr)
        setup_script = '\n'.join(
            line.rstrip()
            for line in [
                '#!/bin/bash',
                'export CANINE_JOB_VARS={}'.format(':'.join(job_vars)),
                'export CANINE_JOB_INPUTS="{}"'.format(os.path.join(compute_env['CANINE_JOBS'], jobId, 'inputs')),
                'export CANINE_JOB_ROOT="{}"'.format(os.path.join(compute_env['CANINE_JOBS'], jobId, 'workspace')),
                'export CANINE_JOB_SETUP="{}"'.format(os.path.join(compute_env['CANINE_JOBS'], jobId, 'setup.sh')),
                'export CANINE_JOB_TEARDOWN="{}"'.format(os.path.join(compute_env['CANINE_JOBS'], jobId, 'teardown.sh')),
            ] + exports + extra_tasks
        ) + '\ncd $CANINE_JOB_ROOT\n'
        teardown_script = '\n'.join(
            line.rstrip()
            for line in [
                '#!/bin/bash',
                'if [[ -d {0} ]]; then cd {0}; fi'.format(os.path.join(compute_env['CANINE_JOBS'], jobId, 'workspace')),
                'if which python3 2>/dev/null >/dev/null; then python3 {0} {1} {2} {3}; else python {0} {1} {2} {3}; fi'.format(
                    os.path.join(compute_env['CANINE_ROOT'], 'delocalization.py'),
                    compute_env['CANINE_OUTPUT'],
                    jobId,
                    ' '.join(
                        '-p {} {}'.format(name, shlex.quote(pattern))
                        for name, pattern in patterns.items()
                    )
                ),
            ]
        )
        return setup_script, teardown_script

    @abc.abstractmethod
    def __enter__(self):
        """
        Enter localizer context.
        May take any setup action required
        """
        pass

    def __exit__(self, *args):
        """
        Exit localizer context
        May take any cleanup action required
        """
        if self._local_dir is not None:
            self._local_dir.cleanup()

    @abc.abstractmethod
    def localize_file(self, src: str, dest: PathType, transport: typing.Optional[AbstractTransport] = None):
        """
        Localizes the given file.
        """
        pass

    @abc.abstractmethod
    def localize(self, inputs: typing.Dict[str, typing.Dict[str, str]], patterns: typing.Dict[str, str], overrides: typing.Optional[typing.Dict[str, typing.Optional[str]]] = None) -> str:
        """
        3 phase task:
        1) Pre-scan inputs to determine proper localization strategy for all inputs
        2) Begin localizing job inputs. For each job, check the predetermined strategy
        and set up the job's setup and teardown scripts
        3) Finally, finalize the localization. This may include broadcasting the
        staging directory or copying a batch of gsutil files
        Returns the remote staging directory, which is now ready for final startup
        """
        pass



#!/usr/bin/python

# Copyright 2017 SchedMD LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import httplib
import os
import shlex
import socket
import subprocess
import time
import urllib
import urllib2
import tempfile

CLUSTER_NAME      = '@CLUSTER_NAME@'
MACHINE_TYPE      = '@MACHINE_TYPE@' # e.g. n1-standard-1, n1-starndard-2
INSTANCE_TYPE     = '@INSTANCE_TYPE@' # e.g. controller, login, compute

PROJECT           = '@PROJECT@'
ZONE              = '@ZONE@'

APPS_DIR          = '/apps'
CURR_SLURM_DIR    = APPS_DIR + '/slurm/current'
MUNGE_DIR         = "/etc/munge"
MUNGE_KEY         = '@MUNGE_KEY@'
SLURM_VERSION     = '@SLURM_VERSION@'
STATIC_NODE_COUNT = @STATIC_NODE_COUNT@
MAX_NODE_COUNT    = @MAX_NODE_COUNT@
DEF_SLURM_ACCT    = '@DEF_SLURM_ACCT@'
DEF_SLURM_USERS   = '@DEF_SLURM_USERS@'
EXTERNAL_COMPUTE_IPS = @EXTERNAL_COMPUTE_IPS@
GPU_TYPE          = '@GPU_TYPE@'
GPU_COUNT         = @GPU_COUNT@
NFS_APPS_SERVER   = '@NFS_APPS_SERVER@'
NFS_APPS_DIR      = '@NFS_APPS_DIR@'
NFS_HOME_SERVER   = '@NFS_HOME_SERVER@'
NFS_HOME_DIR      = '@NFS_HOME_DIR@'
CONTROLLER_SECONDARY_DISK = @CONTROLLER_SECONDARY_DISK@
SEC_DISK_DIR      = '/mnt/disks/sec'
PREEMPTIBLE       = @PREEMPTIBLE@
SUSPEND_TIME      = @SUSPEND_TIME@

DEF_PART_NAME   = "debug"
CONTROL_MACHINE = CLUSTER_NAME + '-controller'

MOTD_HEADER = '''

                                 SSSSSSS
                                SSSSSSSSS
                                SSSSSSSSS
                                SSSSSSSSS
                        SSSS     SSSSSSS     SSSS
                       SSSSSS               SSSSSS
                       SSSSSS    SSSSSSS    SSSSSS
                        SSSS    SSSSSSSSS    SSSS
                SSS             SSSSSSSSS             SSS
               SSSSS    SSSS    SSSSSSSSS    SSSS    SSSSS
                SSS    SSSSSS   SSSSSSSSS   SSSSSS    SSS
                       SSSSSS    SSSSSSS    SSSSSS
                SSS    SSSSSS               SSSSSS    SSS
               SSSSS    SSSS     SSSSSSS     SSSS    SSSSS
          S     SSS             SSSSSSSSS             SSS     S
         SSS            SSSS    SSSSSSSSS    SSSS            SSS
          S     SSS    SSSSSS   SSSSSSSSS   SSSSSS    SSS     S
               SSSSS   SSSSSS   SSSSSSSSS   SSSSSS   SSSSS
          S    SSSSS    SSSS     SSSSSSS     SSSS    SSSSS    S
    S    SSS    SSS                                   SSS    SSS    S
    S     S                                                   S     S
                SSS
                SSS
                SSS
                SSS
 SSSSSSSSSSSS   SSS   SSSS       SSSS    SSSSSSSSS   SSSSSSSSSSSSSSSSSSSS
SSSSSSSSSSSSS   SSS   SSSS       SSSS   SSSSSSSSSS  SSSSSSSSSSSSSSSSSSSSSS
SSSS            SSS   SSSS       SSSS   SSSS        SSSS     SSSS     SSSS
SSSS            SSS   SSSS       SSSS   SSSS        SSSS     SSSS     SSSS
SSSSSSSSSSSS    SSS   SSSS       SSSS   SSSS        SSSS     SSSS     SSSS
 SSSSSSSSSSSS   SSS   SSSS       SSSS   SSSS        SSSS     SSSS     SSSS
         SSSS   SSS   SSSS       SSSS   SSSS        SSSS     SSSS     SSSS
         SSSS   SSS   SSSS       SSSS   SSSS        SSSS     SSSS     SSSS
SSSSSSSSSSSSS   SSS   SSSSSSSSSSSSSSS   SSSS        SSSS     SSSS     SSSS
SSSSSSSSSSSS    SSS    SSSSSSSSSSSSS    SSSS        SSSS     SSSS     SSSS


'''

def add_slurm_user():

    SLURM_UID = str(992)
    subprocess.call(['groupadd', '-g', SLURM_UID, 'slurm'])
    subprocess.call(['useradd', '-m', '-c', 'SLURM Workload Manager',
        '-d', '/var/lib/slurm', '-u', SLURM_UID, '-g', 'slurm',
        '-s', '/bin/bash', 'slurm'])

# END add_slurm_user()


def setup_modules():

    appsmfs = '/apps/modulefiles'

    if appsmfs not in open('/usr/share/Modules/init/.modulespath').read():
        if INSTANCE_TYPE == 'controller' and not os.path.isdir(appsmfs):
            subprocess.call(['mkdir', '-p', appsmfs])

        with open('/usr/share/Modules/init/.modulespath', 'a') as dotmp:
            dotmp.write(appsmfs)

# END setup_modules


def start_motd():

    msg = MOTD_HEADER + """
*** Slurm is currently being installed/configured in the background. ***
A terminal broadcast will announce when installation and configuration is
complete.

Partition {} will be marked down until the compute image has been created.
For instances with gpus attached, it could take ~10 mins after the controller
has finished installing.

""".format(DEF_PART_NAME)

    if INSTANCE_TYPE != "controller":
        msg += """/home on the controller will be mounted over the existing /home.
Any changes in /home will be hidden. Please wait until the installation is
complete before making changes in your home directory.

"""

    f = open('/etc/motd', 'w')
    f.write(msg)
    f.close()

# END start_motd()


def end_motd(broadcast=True):

    f = open('/etc/motd', 'w')
    f.write(MOTD_HEADER)
    f.close()

    if not broadcast:
        return

    subprocess.call(['wall', '-n',
        '*** Slurm ' + INSTANCE_TYPE + ' daemon installation complete ***'])

    if INSTANCE_TYPE != "controller":
        subprocess.call(['wall', '-n', """
/home on the controller was mounted over the existing /home.
Either log out and log back in or cd into ~.
"""])

#END start_motd()


def have_internet():
    conn = httplib.HTTPConnection("www.google.com", timeout=1)
    try:
        conn.request("HEAD", "/")
        conn.close()
        return True
    except:
        conn.close()
        return False

#END have_internet()


def install_packages():

    packages = ['bind-utils',
                'epel-release',
                'gcc',
                'git',
                'hwloc',
                'hwloc-devel',
                'libibmad',
                'libibumad',
                'lua',
                'lua-devel',
                'man2html',
                'mariadb',
                'mariadb-devel',
                'mariadb-server',
                'munge',
                'munge-devel',
                'munge-libs',
                'ncurses-devel',
                'nfs-utils',
                'numactl',
                'numactl-devel',
                'openssl-devel',
                'pam-devel',
                'perl-ExtUtils-MakeMaker',
                'python-pip',
                'readline-devel',
                'rpm-build',
                'rrdtool-devel',
                'vim',
                'wget',
                'tmux',
                'pdsh',
                'openmpi'
               ]

    while subprocess.call(['yum', 'install', '-y'] + packages):
        print "yum failed to install packages. Trying again in 5 seconds"
        time.sleep(5)

    while subprocess.call(['pip', 'install', '--upgrade',
        'google-api-python-client']):
        print "failed to install google python api client. Trying again 5 seconds."
        time.sleep(5)

    if GPU_COUNT and (INSTANCE_TYPE == "compute"):
        rpm = "cuda-repo-rhel7-10.0.130-1.x86_64.rpm"
        subprocess.call("yum -y install kernel-devel-$(uname -r) kernel-headers-$(uname -r)", shell=True)
        subprocess.call(shlex.split("wget http://developer.download.nvidia.com/compute/cuda/repos/rhel7/x86_64/" + rpm))
        subprocess.call(shlex.split("rpm -i " + rpm))
        subprocess.call(shlex.split("yum clean all"))
        subprocess.call(shlex.split("yum -y install cuda"))
        subprocess.call(shlex.split("nvidia-smi")) # Creates the device files

#END install_packages()


def setup_munge():

    munge_service_patch = "/usr/lib/systemd/system/munge.service"
    f = open(munge_service_patch, 'w')
    f.write("""
[Unit]
Description=MUNGE authentication service
Documentation=man:munged(8)
After=network.target
After=syslog.target
After=time-sync.target
""")

    if (INSTANCE_TYPE != "controller"):
        f.write("RequiresMountsFor={}\n".format(MUNGE_DIR))

    f.write("""
[Service]
Type=forking
ExecStart=/usr/sbin/munged --num-threads=10
PIDFile=/var/run/munge/munged.pid
User=munge
Group=munge
Restart=on-abort

[Install]
WantedBy=multi-user.target""")
    f.close()

    subprocess.call(['systemctl', 'enable', 'munge'])

    if (INSTANCE_TYPE != "controller"):
        f = open('/etc/fstab', 'a')
        f.write("""
{1}:{0}    {0}     nfs      rw,hard,intr  0     0
""".format(MUNGE_DIR, CONTROL_MACHINE))
        f.close()
        return

    if MUNGE_KEY:
        f = open(MUNGE_DIR +'/munge.key', 'w')
        f.write(MUNGE_KEY)
        f.close()

        subprocess.call(['chown', '-R', 'munge:', MUNGE_DIR, '/var/log/munge/'])
        os.chmod(MUNGE_DIR + '/munge.key' ,0o400)
        os.chmod(MUNGE_DIR                ,0o700)
        os.chmod('/var/log/munge/'        ,0o700)
    else:
        subprocess.call(['create-munge-key'])

#END setup_munge ()

def start_munge():
        subprocess.call(['systemctl', 'start', 'munge'])
#END start_munge()

def setup_nfs_exports():

    f = open('/etc/exports', 'w')
    if not NFS_HOME_SERVER:
        f.write("""
/home  *(rw,no_subtree_check,no_root_squash)
""")
    if not NFS_APPS_SERVER:
        f.write("""
%s  *(rw,no_subtree_check,no_root_squash)
""" % APPS_DIR)
    f.write("""
/etc/munge *(rw,no_subtree_check,no_root_squash)
""")
    if CONTROLLER_SECONDARY_DISK:
        f.write("""
%s  *(rw,no_subtree_check,no_root_squash)
""" % SEC_DISK_DIR)
    f.close()

    subprocess.call(shlex.split("exportfs -a"))

#END setup_nfs_exports()


def expand_machine_type():

    # Force re-evaluation of site-packages so that namespace packages (such
    # as google-auth) are importable. This is needed because we install the
    # packages while this script is running and do not have the benefit of
    # restarting the interpreter for it to do it's usual startup sequence to
    # configure import magic.
    import sys
    import site
    for path in [x for x in sys.path if 'site-packages' in x]:
        site.addsitedir(path)

    import googleapiclient.discovery

    # Assume sockets is 1. Currently, no instances with multiple sockets
    # Assume hyper-threading is on and 2 threads per core
    machine = {'sockets': 1, 'cores': 1, 'threads': 1, 'memory': 1}

    try:
        compute = googleapiclient.discovery.build('compute', 'v1',
                                                  cache_discovery=False)
        type_resp = compute.machineTypes().get(project=PROJECT, zone=ZONE,
                machineType=MACHINE_TYPE).execute()
        if type_resp:
            tot_cpus = type_resp['guestCpus']
            if tot_cpus > 1:
                # machine['cores']   = tot_cpus / 2
                # machine['threads'] = 2
                # Allocate threads as individual cores because hyper-threading
                machine['cores'] = tot_cpus

            # Because the actual memory on the host will be different than what
            # is configured (e.g. kernel will take it). From experiments, about
            # 16 MB per GB are used (plus about 400 MB buffer for the first
            # couple of GB's. Using 30 MB to be safe.
            gb = type_resp['memoryMb'] / 1024;
            machine['memory'] = type_resp['memoryMb'] - (400 + (gb * 30))

    except Exception, e:
        print "Failed to get MachineType '%s' from google api (%s)" % (MACHINE_TYPE, str(e))

    return machine
#END expand_machine_type()


def install_slurm_conf():

    machine = expand_machine_type()
    def_mem_per_cpu = max(100,
            (machine['memory'] /
             (machine['threads']*machine['cores']*machine['sockets'])))

    conf = """
# slurm.conf file generated by configurator.html.
# Put this file on all nodes of your cluster.
# See the slurm.conf man page for more information.
#
ControlMachine={control_machine}
#ControlAddr=
#BackupController=
#BackupAddr=
#
AuthType=auth/munge
AuthInfo=cred_expire=120
#CheckpointType=checkpoint/none
CryptoType=crypto/munge
#DisableRootJobs=NO
#EnforcePartLimits=NO
#Epilog=
#EpilogSlurmctld=
#FirstJobId=1
#MaxJobId=999999
#GroupUpdateForce=0
#GroupUpdateTime=600
#JobCheckpointDir=/var/slurm/checkpoint
#JobCredentialPrivateKey=
#JobCredentialPublicCertificate=
#JobFileAppend=0
#JobRequeue=1
#JobSubmitPlugins=1
#KillOnBadExit=0
#LaunchType=launch/slurm
#Licenses=foo*4,bar
#MailProg=/bin/mail
#MaxJobCount=5000
#MaxStepCount=40000
#MaxTasksPerNode=128
MpiDefault=none
#MpiParams=ports=#-#
#PluginDir=
#PlugStackConfig=
#PrivateData=jobs
LaunchParameters=send_gids

# Always show cloud nodes. Otherwise cloud nodes are hidden until they are
# resumed. Having them shown can be useful in detecting downed nodes.
# NOTE: slurm won't allocate/resume nodes that are down. So in the case of
# preemptible nodes -- if gcp preempts a node, the node will eventually be put
# into a down date because the node will stop responding to the controller.
# (e.g. SlurmdTimeout).
PrivateData=cloud

ProctrackType=proctrack/cgroup

#Prolog=
#PrologFlags=
#PrologSlurmctld=
#PropagatePrioProcess=0
#PropagateResourceLimits=
#PropagateResourceLimitsExcept=Sched
#RebootProgram=

ReturnToService=2
#SallocDefaultCommand=
SlurmctldPidFile=/var/run/slurm/slurmctld.pid
SlurmctldPort=6820-6830
SlurmdPidFile=/var/run/slurm/slurmd.pid
SlurmdPort=6818
SlurmdSpoolDir=/var/spool/slurmd
SlurmUser=slurm
#SlurmdUser=root
#SrunEpilog=
#SrunProlog=
StateSaveLocation={apps_dir}/slurm/state
SwitchType=switch/none
#TaskEpilog=
TaskPlugin=task/affinity,task/cgroup
#TaskPluginParam=
#TaskProlog=
#TopologyPlugin=topology/tree
#TmpFS=/tmp
#TrackWCKey=no
#TreeWidth=
#UnkillableStepProgram=
#UsePAM=0
#
#
# TIMERS
#BatchStartTimeout=10
#CompleteWait=0
#EpilogMsgTime=2000
#GetEnvTimeout=2
#HealthCheckInterval=0
#HealthCheckProgram=
InactiveLimit=0
KillWait=30
MessageTimeout=60
#ResvOverRun=0
MinJobAge=300
#OverTimeLimit=0
SlurmctldTimeout=120
SlurmdTimeout=300
#UnkillableStepTimeout=60
#VSizeFactor=0
Waittime=0
#
#
# SCHEDULING
FastSchedule=1
DefMemPerCPU={def_mem_per_cpu}
#MaxMemPerCPU=0
#SchedulerTimeSlice=30
SchedulerType=sched/backfill
SelectType=select/cons_res
SelectTypeParameters=CR_CPU_Memory
#
#
# JOB PRIORITY
#PriorityFlags=
#PriorityType=priority/basic
#PriorityDecayHalfLife=
#PriorityCalcPeriod=
#PriorityFavorSmall=
#PriorityMaxAge=
#PriorityUsageResetPeriod=
#PriorityWeightAge=
#PriorityWeightFairshare=
#PriorityWeightJobSize=
#PriorityWeightPartition=
#PriorityWeightQOS=
#
#
# LOGGING AND ACCOUNTING
AccountingStorageEnforce=associations,limits,qos,safe
AccountingStorageHost={control_machine}
#AccountingStorageLoc=
#AccountingStoragePass=
#AccountingStoragePort=
AccountingStorageType=accounting_storage/slurmdbd
#AccountingStorageUser=
AccountingStoreJobComment=YES
ClusterName={cluster_name}
#DebugFlags=powersave
#JobCompHost=
#JobCompLoc=
#JobCompPass=
#JobCompPort=
JobCompType=jobcomp/none
#JobCompUser=
#JobContainerType=job_container/none
JobAcctGatherFrequency=30
JobAcctGatherType=jobacct_gather/linux
SlurmctldDebug=info
SlurmctldLogFile={apps_dir}/slurm/log/slurmctld.log
SlurmdDebug=debug
SlurmdLogFile=/var/log/slurm/slurmd-%n.log
#
#
# POWER SAVE SUPPORT FOR IDLE NODES (optional)
SuspendProgram={apps_dir}/slurm/scripts/suspend.py
ResumeProgram={apps_dir}/slurm/scripts/resume.py
ResumeFailProgram={apps_dir}/slurm/scripts/suspend.py
SuspendTimeout=600
ResumeTimeout=1200
ResumeRate=0
#SuspendExcNodes=
#SuspendExcParts=
SuspendRate=0
SuspendTime={suspend_time}
#
SchedulerParameters=salloc_wait_nodes
SlurmctldParameters=cloud_dns,idle_on_node_suspend
CommunicationParameters=NoAddrCache
#
# COMPUTE NODES
""".format(apps_dir        = APPS_DIR,
           cluster_name    = CLUSTER_NAME,
           control_machine = CONTROL_MACHINE,
           def_mem_per_cpu = def_mem_per_cpu,
           suspend_time    = SUSPEND_TIME)

    if GPU_COUNT:
        conf += "GresTypes=gpu\n"

    conf += ' '.join(("NodeName=DEFAULT",
                      # "Sockets="        + str(machine['sockets']),
                      "CPUs=" + str(machine['cores']),
                      # "ThreadsPerCore=" + str(machine['threads']),
                      "RealMemory="     + str(machine['memory']),
                      "State=UNKNOWN"))

    if GPU_COUNT:
        conf += " Gres=gpu:" + str(GPU_COUNT)
    conf += "\n"

    static_range = ""
    if STATIC_NODE_COUNT and STATIC_NODE_COUNT > 1:
        static_range = "[1-%d]" % STATIC_NODE_COUNT
    elif STATIC_NODE_COUNT:
        static_range = "1"

    cloud_range = ""
    if MAX_NODE_COUNT and (MAX_NODE_COUNT != STATIC_NODE_COUNT):
        cloud_range = "[%d-%d]" % (STATIC_NODE_COUNT+1, MAX_NODE_COUNT)

    if static_range:
        conf += """
SuspendExcNodes={1}-compute{0}
NodeName={1}-compute{0}
""".format(static_range, CLUSTER_NAME)

    if cloud_range:
        conf += "NodeName={0}-compute{1} State=CLOUD".format(CLUSTER_NAME, cloud_range)

    conf += """
PartitionName={} Nodes={}-compute[1-{}] Default=YES MaxTime=INFINITE State=UP LLN=yes
""".format(DEF_PART_NAME, CLUSTER_NAME, MAX_NODE_COUNT)

    etc_dir = CURR_SLURM_DIR + '/etc'
    if not os.path.exists(etc_dir):
        os.makedirs(etc_dir)
    f = open(etc_dir + '/slurm.conf', 'w')
    f.write(conf)
    f.close()
#END install_slurm_conf()


def install_slurmdbd_conf():

    conf = """
#ArchiveEvents=yes
#ArchiveJobs=yes
#ArchiveResvs=yes
#ArchiveSteps=no
#ArchiveSuspend=no
#ArchiveTXN=no
#ArchiveUsage=no

AuthType=auth/munge
DbdHost={control_machine}
DebugLevel=debug2

#PurgeEventAfter=1month
#PurgeJobAfter=12month
#PurgeResvAfter=1month
#PurgeStepAfter=1month
#PurgeSuspendAfter=1month
#PurgeTXNAfter=12month
#PurgeUsageAfter=24month

LogFile={apps_dir}/slurm/log/slurmdbd.log
PidFile=/var/run/slurm/slurmdbd.pid

SlurmUser=slurm
StorageUser=slurm

StorageLoc=slurm_acct_db

StorageType=accounting_storage/mysql
#StorageUser=database_mgr
#StoragePass=shazaam

""".format(apps_dir = APPS_DIR, control_machine = CONTROL_MACHINE)
    etc_dir = CURR_SLURM_DIR + '/etc'
    if not os.path.exists(etc_dir):
        os.makedirs(etc_dir)
    f = open(etc_dir + '/slurmdbd.conf', 'w')
    f.write(conf)
    f.close()

#END install_slurmdbd_conf()


def install_cgroup_conf():

    conf = """
CgroupAutomount=no
#CgroupMountpoint=/sys/fs/cgroup
ConstrainCores=yes
ConstrainRamSpace=yes
ConstrainSwapSpace=yes
TaskAffinity=no
ConstrainDevices=yes
"""

    etc_dir = CURR_SLURM_DIR + '/etc'
    f = open(etc_dir + '/cgroup.conf', 'w')
    f.write(conf)
    f.close()

    f = open(etc_dir + '/cgroup_allowed_devices_file.conf', 'w')
    f.write("")
    f.close()

    if GPU_COUNT:
        f = open(etc_dir + '/gres.conf', 'w')
        f.write("NodeName=%s-compute[1-%d] Name=gpu File=/dev/nvidia[0-%d]"
                % (CLUSTER_NAME, MAX_NODE_COUNT, (GPU_COUNT - 1)))
        f.close()
#END install_cgroup_conf()


def install_meta_files():

    scripts_path = APPS_DIR + '/slurm/scripts'
    if not os.path.exists(scripts_path):
        os.makedirs(scripts_path)

    GOOGLE_URL = "http://metadata.google.internal/computeMetadata/v1/instance/attributes"

    meta_files = [
        {'file': 'suspend.py', 'meta': 'slurm_suspend'},
        {'file': 'resume.py', 'meta': 'slurm_resume'},
        {'file': 'startup-script.py', 'meta': 'startup-script-compute'},
        {'file': 'slurm-gcp-sync.py', 'meta': 'slurm-gcp-sync'},
        {'file': 'compute-shutdown', 'meta': 'compute-shutdown'},
        {'file': 'custom-compute-install', 'meta': 'custom-compute-install'},
        {'file': 'custom-controller-install', 'meta': 'custom-controller-install'},
    ]

    for meta in meta_files:
        file_name = meta['file']
        meta_name = meta['meta']

        req = urllib2.Request("{}/{}".format(GOOGLE_URL, meta_name))
        req.add_header('Metadata-Flavor', 'Google')
        resp = urllib2.urlopen(req)

        f = open("{}/{}".format(scripts_path, file_name), 'w')
        f.write(resp.read())
        f.close()
        os.chmod("{}/{}".format(scripts_path, file_name), 0o755)

        subprocess.call(shlex.split("gcloud compute instances remove-metadata {} --zone={} --keys={}".
                                    format(CONTROL_MACHINE, ZONE, meta_name)))

#END install_meta_files()

def install_slurm():

    SLURM_PREFIX = "";

    prev_path = os.getcwd()

    SRC_PATH = APPS_DIR + "/slurm/src"
    if not os.path.exists(SRC_PATH):
        os.makedirs(SRC_PATH)
    os.chdir(SRC_PATH)

    use_version = "";
    if (SLURM_VERSION[0:2] == "b:"):
        GIT_URL = "https://github.com/SchedMD/slurm.git"
        use_version = SLURM_VERSION[2:]
        subprocess.call(
            shlex.split("git clone -b {0} {1} {0}".format(
                use_version, GIT_URL)))
    else:
        SCHEDMD_URL = 'https://download.schedmd.com/slurm/'
        file = "slurm-%s.tar.bz2" % SLURM_VERSION
        urllib.urlretrieve(SCHEDMD_URL + file, SRC_PATH + '/' + file)

        cmd = "tar -xvjf " + file
        use_version = subprocess.check_output(
            shlex.split(cmd)).splitlines()[0][:-1]

    os.chdir(use_version)
    SLURM_PREFIX  = APPS_DIR + '/slurm/' + use_version

    if not os.path.exists('build'):
        os.makedirs('build')
    os.chdir('build')
    subprocess.call(['../configure', '--prefix=%s' % SLURM_PREFIX,
                     '--sysconfdir=%s/etc' % CURR_SLURM_DIR])
    subprocess.call(['make', '-j', 'install'])

    subprocess.call(shlex.split("ln -s %s %s" % (SLURM_PREFIX, CURR_SLURM_DIR)))

    os.chdir(prev_path)

    if not os.path.exists(APPS_DIR + '/slurm/state'):
        os.makedirs(APPS_DIR + '/slurm/state')
        subprocess.call(['chown', '-R', 'slurm:', APPS_DIR + '/slurm/state'])
    if not os.path.exists(APPS_DIR + '/slurm/log'):
        os.makedirs(APPS_DIR + '/slurm/log')
        subprocess.call(['chown', '-R', 'slurm:', APPS_DIR + '/slurm/log'])

    install_slurm_conf()
    install_slurmdbd_conf()
    install_cgroup_conf()
    install_meta_files()

#END install_slurm()

def install_slurm_tmpfile():

    run_dir = '/var/run/slurm'

    f = open('/etc/tmpfiles.d/slurm.conf', 'w')
    f.write("""
d %s 0755 slurm slurm -
""" % run_dir)
    f.close()

    if not os.path.exists(run_dir):
        os.makedirs(run_dir)

    os.chmod(run_dir, 0o755)
    subprocess.call(['chown', 'slurm:', run_dir])

#END install_slurm_tmpfile()

def install_controller_service_scripts():

    install_slurm_tmpfile()

    # slurmctld.service
    f = open('/usr/lib/systemd/system/slurmctld.service', 'w')
    f.write("""
[Unit]
Description=Slurm controller daemon
After=network.target munge.service
ConditionPathExists={prefix}/etc/slurm.conf

[Service]
Type=forking
EnvironmentFile=-/etc/sysconfig/slurmctld
ExecStart={prefix}/sbin/slurmctld $SLURMCTLD_OPTIONS
ExecReload=/bin/kill -HUP $MAINPID
PIDFile=/var/run/slurm/slurmctld.pid

[Install]
WantedBy=multi-user.target
""".format(prefix = CURR_SLURM_DIR))
    f.close()

    os.chmod('/usr/lib/systemd/system/slurmctld.service', 0o644)

    # slurmdbd.service
    f = open('/usr/lib/systemd/system/slurmdbd.service', 'w')
    f.write("""
[Unit]
Description=Slurm DBD accounting daemon
After=network.target munge.service
ConditionPathExists={prefix}/etc/slurmdbd.conf

[Service]
Type=forking
EnvironmentFile=-/etc/sysconfig/slurmdbd
ExecStart={prefix}/sbin/slurmdbd $SLURMDBD_OPTIONS
ExecReload=/bin/kill -HUP $MAINPID
PIDFile=/var/run/slurm/slurmdbd.pid

[Install]
WantedBy=multi-user.target
""".format(prefix = CURR_SLURM_DIR))
    f.close()

    os.chmod('/usr/lib/systemd/system/slurmdbd.service', 0o644)

#END install_controller_service_scripts()


def install_compute_service_scripts():

    install_slurm_tmpfile()

    # slurmd.service
    f = open('/usr/lib/systemd/system/slurmd.service', 'w')
    f.write("""
[Unit]
Description=Slurm node daemon
After=network.target munge.service
ConditionPathExists={prefix}/etc/slurm.conf

[Service]
Type=forking
EnvironmentFile=-/etc/sysconfig/slurmd
ExecStart={prefix}/sbin/slurmd $SLURMD_OPTIONS
ExecReload=/bin/kill -HUP $MAINPID
PIDFile=/var/run/slurm/slurmd.pid
KillMode=process
LimitNOFILE=51200
LimitMEMLOCK=infinity
LimitSTACK=infinity

[Install]
WantedBy=multi-user.target
""".format(prefix = CURR_SLURM_DIR))
    f.close()

    os.chmod('/usr/lib/systemd/system/slurmd.service', 0o644)
    subprocess.call(shlex.split('systemctl enable slurmd'))

#END install_compute_service_scripts()


def setup_bash_profile():

    f = open('/etc/profile.d/slurm.sh', 'w')
    f.write("""
S_PATH=%s
PATH=$PATH:$S_PATH/bin:$S_PATH/sbin
""" % CURR_SLURM_DIR)
    f.close()

    if GPU_COUNT and (INSTANCE_TYPE == "compute"):
        f = open('/etc/profile.d/cuda.sh', 'w')
        f.write("""
CUDA_PATH=/usr/local/cuda
PATH=$CUDA_PATH/bin${PATH:+:${PATH}}
LD_LIBRARY_PATH=$CUDA_PATH/lib64${LD_LIBRARY_PATH:+:${LD_LIBRARY_PATH}}
""")
        f.close()

#END setup_bash_profile()

def setup_nfs_apps_vols():

    f = open('/etc/fstab', 'a')
    if not NFS_APPS_SERVER:
        if ((INSTANCE_TYPE != "controller")):
            f.write("""
{1}:{0}    {0}     nfs      rw,hard,intr  0     0
""".format(APPS_DIR, CONTROL_MACHINE))
    else:
        f.write("""
{1}:{2}    {0}     nfs      rw,hard,intr  0     0
""".format(APPS_DIR, NFS_APPS_SERVER, NFS_APPS_DIR))
    f.close()

#END setup_nfs_apps_vols()

def setup_nfs_home_vols():

    f = open('/etc/fstab', 'a')
    if not NFS_HOME_SERVER:
        if ((INSTANCE_TYPE != "controller")):
            f.write("""
{0}:/home    /home     nfs      rw,hard,intr  0     0
""".format(CONTROL_MACHINE))
    else:
        f.write("""
{0}:{1}    /home     nfs      rw,hard,intr  0     0
""".format(NFS_HOME_SERVER, NFS_HOME_DIR))
    f.close()

#END setup_nfs_home_vols()

def setup_nfs_sec_vols():
    f = open('/etc/fstab', 'a')

    if CONTROLLER_SECONDARY_DISK:
        if ((INSTANCE_TYPE != "controller")):
            f.write("""
{1}:{0}    {0}     nfs      rw,hard,intr  0     0
""".format(SEC_DISK_DIR, CONTROL_MACHINE))
    f.close()

#END setup_nfs_sec_vols()

def setup_secondary_disks():

    subprocess.call(shlex.split("sudo mkfs.ext4 -m 0 -F -E lazy_itable_init=0,lazy_journal_init=0,discard /dev/sdb"))
    f = open('/etc/fstab', 'a')

    f.write("""
/dev/sdb    {0}  ext4    discard,defaults,nofail  0  2
""".format(SEC_DISK_DIR))
    f.close()

#END setup_secondary_disks()

def mount_nfs_vols():
    while subprocess.call(['mount', '-a']):
        print "Waiting for " + APPS_DIR + " and /home to be mounted"
        time.sleep(5)

#END mount_nfs_vols()

# Tune the NFS server to support many mounts
def setup_nfs_threads():

    f = open('/etc/sysconfig/nfs', 'a')
    f.write("""
# Added by Google
RPCNFSDCOUNT=256
""".format(APPS_DIR))
    f.close()

# END setup_nfs_threads()

def setup_sync_cronjob():

    os.system("echo '*/1 * * * * {}/slurm/scripts/slurm-gcp-sync.py' | crontab -u root -".format(APPS_DIR))

# END setup_sync_cronjob()

def setup_slurmd_cronjob():
    #subprocess.call(shlex.split('crontab < /apps/slurm/scripts/cron'))
    os.system("echo '*/2 * * * * if [ `systemctl status slurmd | grep -c inactive` -gt 0 ]; then mount -a; systemctl restart slurmd; fi' | crontab -u root -")
# END setup_slurmd_cronjob()

def create_compute_image():

    end_motd(False)
    subprocess.call("sync")
    ver = datetime.datetime.now().strftime("%Y-%m-%d-%H-%M-%S")

    if GPU_COUNT:
        time.sleep(300)

    print "Creating compute image..."
    # STARTUP_SCRIPT = "'#!/bin/bash'$'\\n''{slurm_dir}/bin/scontrol update partitionname={part} state=up'$'\\n''gcloud compute instances delete --zone {zone} {name}-cluster-starter --delete-disks boot --quiet'".format(
    #     slurm_dir=CURR_SLURM_DIR,
    #     part=DEF_PART_NAME,
    #     zone=ZONE,
    #     name=CLUSTER_NAME
    # )
    # 1) Create an initial image. Likely will work, but may be broken depending on disk state
    subprocess.call(shlex.split(
        "gcloud compute images "
        "create {0}-compute-image-{3} "
        "--source-disk {1} "
        "--source-disk-zone {2} --force "
        "--family {0}-compute-image-family\n".format(
            CLUSTER_NAME, socket.gethostname(), ZONE, ver)
    ))
    # 2) Boot a new VM which will stop this one and create a new image
    with tempfile.NamedTemporaryFile('w') as w:
        w.write('#!/bin/bash\n')
        # 2.a) Stop the compute-image vm
        w.write(
            'gcloud compute instances stop --zone {zone} {name}-compute-image\n'.format(
                zone=ZONE,
                name=CLUSTER_NAME
            )
        )
        # 2.b) Create the final image
        w.write(
            "gcloud compute images "
            "create {0}-compute-image-{3}-final "
            "--source-disk {1} "
            "--source-disk-zone {2} --force "
            "--family {0}-compute-image-family\n".format(
                CLUSTER_NAME, socket.gethostname(), ZONE, ver)
        )
        # 2.c) Delete the original image
        w.write(
            "gcloud compute images "
            "delete {}-compute-image-{}\n".format(
                CLUSTER_NAME, ver)
        )
        # w.write(
        #     'gcloud compute instances create {name}-cluster-starter '
        #     '--zone {zone} --image {name}-compute-image-{version} '
        #     '--boot-disk-type pd-standard --machine-type f1-micro --async '
        #     "--metadata-from-file startup-script=<(echo {startup_script}) "
        #     '--network-interface subnet=projects/{project}/regions/{region}/subnetworks/default '
        #     '--scopes https://www.googleapis.com/auth/cloud-platform\n'.format(
        #         name=CLUSTER_NAME,
        #         zone=ZONE,
        #         version=ver,
        #         startup_script=STARTUP_SCRIPT,
        #         project=PROJECT,
        #         region=ZONE[:-2], #strip "-[a-z]" from end of zone
        #     )
        # )
        # 2.d) Delete the compute-image vm
        w.write(
            'gcloud compute instances delete --zone {zone} {name}-compute-image --delete-disks boot --quiet\n'.format(
                zone=ZONE,
                name=CLUSTER_NAME
            )
        )
        # 2.e) Delete this VM
        w.write(
            'gcloud compute instances delete --zone {zone} {name}-image-taker --delete-disks boot --quiet\n'.format(
                zone=ZONE,
                name=CLUSTER_NAME
            )
        )
        w.flush()
        # 3) Mark partition up. If we're lucky, the initial image will work
        # If we're not, the final image will be ready shortly
        subprocess.call(shlex.split(
            '{slurm_dir}/bin/scontrol update partitionname={part} state=up'.format(
                slurm_dir=CURR_SLURM_DIR,
                part=DEF_PART_NAME,
            )
        ))
        # 4) Start that image-taker vm
        subprocess.call(shlex.split(
            'gcloud compute instances create {name}-image-taker '
            '--zone {zone} --image-family ubuntu-1604-lts --image-project ubuntu-os-cloud '
            '--boot-disk-type pd-standard --machine-type f1-micro '
            '--metadata-from-file startup-script={script_path} '
            '--scopes https://www.googleapis.com/auth/cloud-platform\n'.format(
                name=CLUSTER_NAME,
                zone=ZONE,
                script_path=w.name
            )
        ))


#END create_compute_image()


def setup_selinux():

    subprocess.call(shlex.split('setenforce 0'))
    f = open('/etc/selinux/config', 'w')
    f.write("""
SELINUX=permissive
SELINUXTYPE=targeted
""")
    f.close()
#END setup_selinux()


def main():

    hostname = socket.gethostname()

    setup_selinux()

    if INSTANCE_TYPE == "compute":
        while not have_internet():
            print "Waiting for internet connection"

    if not os.path.exists(APPS_DIR + '/slurm'):
        os.makedirs(APPS_DIR + '/slurm')
        print "ww Created Slurm Folders"

    if CONTROLLER_SECONDARY_DISK:
        if not os.path.exists(SEC_DISK_DIR):
            os.makedirs(SEC_DISK_DIR)

    start_motd()

    if not os.path.exists('/var/log/slurm'):
        os.makedirs('/var/log/slurm')

    add_slurm_user()
    install_packages()
    setup_munge()
    setup_bash_profile()

    if (CONTROLLER_SECONDARY_DISK and (INSTANCE_TYPE == "controller")):
        setup_secondary_disks()

    setup_nfs_apps_vols()
    setup_nfs_home_vols()
    setup_nfs_sec_vols()

    if INSTANCE_TYPE == "controller":
        mount_nfs_vols()
        setup_modules()
        start_munge()
        install_slurm()

        try:
            subprocess.call("sudo {}/slurm/scripts/custom-controller-install 2>&1 > ~/install.log"
                            .format(APPS_DIR), shell=True, executable='/bin/bash')
        except Exception:
            # Ignore blank files with no shell magic.
            pass

        install_controller_service_scripts()

        subprocess.call(shlex.split('systemctl enable mariadb'))
        subprocess.call(shlex.split('systemctl start mariadb'))

        subprocess.call(['mysql', '-u', 'root', '-e',
            "create user 'slurm'@'localhost'"])
        subprocess.call(['mysql', '-u', 'root', '-e',
            "grant all on slurm_acct_db.* TO 'slurm'@'localhost';"])
        subprocess.call(['mysql', '-u', 'root', '-e',
            "grant all on slurm_acct_db.* TO 'slurm'@'{0}';".format(CONTROL_MACHINE)])

        subprocess.call(shlex.split('systemctl enable slurmdbd'))
        subprocess.call(shlex.split('systemctl start slurmdbd'))

        # Wait for slurmdbd to come up
        time.sleep(5)

        oslogin_chars = ['@', '.']

        SLURM_USERS = DEF_SLURM_USERS

        for char in oslogin_chars:
            SLURM_USERS = SLURM_USERS.replace(char, '_')

        subprocess.call(shlex.split(CURR_SLURM_DIR + '/bin/sacctmgr -i add cluster ' + CLUSTER_NAME))
        subprocess.call(shlex.split(CURR_SLURM_DIR + '/bin/sacctmgr -i add account ' + DEF_SLURM_ACCT))
        subprocess.call(shlex.split(CURR_SLURM_DIR + '/bin/sacctmgr -i add user ' + SLURM_USERS + ' account=' + DEF_SLURM_ACCT))

        subprocess.call(shlex.split('systemctl enable slurmctld'))
        subprocess.call(shlex.split('systemctl start slurmctld'))
        setup_nfs_threads()
        # Export at the end to signal that everything is up
        subprocess.call(shlex.split('systemctl enable nfs-server'))
        subprocess.call(shlex.split('systemctl start nfs-server'))
        setup_nfs_exports()

        if PREEMPTIBLE:
            setup_sync_cronjob()

        # DOWN partition until image is created.
        subprocess.call(shlex.split(
            "{}/bin/scontrol update partitionname={} state=down".format(
                CURR_SLURM_DIR, DEF_PART_NAME)))

        print "ww Done installing controller"
    elif INSTANCE_TYPE == "compute":
        install_compute_service_scripts()
        setup_slurmd_cronjob()
        mount_nfs_vols()
        start_munge()

        try:
            subprocess.call("sudo {}/slurm/scripts/custom-compute-install 2>&1 > ~/install.log"
                            .format(APPS_DIR), shell=True, executable='/bin/bash')
        except Exception:
            # Ignore blank files with no shell magic.
            pass

        if hostname == CLUSTER_NAME + "-compute-image":
            create_compute_image()
        else:
            subprocess.call(shlex.split('systemctl start slurmd'))

    else: # login nodes
        mount_nfs_vols()
        start_munge()

        try:
            subprocess.call("{}/slurm/scripts/custom-compute-install"
                            .format(APPS_DIR))
        except Exception:
            # Ignore blank files with no shell magic.
            pass


    if hostname != CLUSTER_NAME + "-compute-image":
        # Wait for the compute image to mark the partition up
        part_state = subprocess.check_output(shlex.split(
            "{}/bin/scontrol show part {}".format(
                CURR_SLURM_DIR, DEF_PART_NAME)))
        while "State=UP" not in part_state:
            part_state = subprocess.check_output(shlex.split(
                "{}/bin/scontrol show part {}".format(
                    CURR_SLURM_DIR, DEF_PART_NAME)))

    end_motd()

    subprocess.call(shlex.split("gcloud compute instances remove-metadata {} "
                                "--zone={} --keys=startup-script"
                                .format(hostname, ZONE)))
# END main()


if __name__ == '__main__':
    main()



import getpass
import tempfile
import typing
import time
import subprocess
import shutil
import os
import sys
from .remote import RemoteSlurmBackend
from ..utils import get_default_gcp_project, ArgumentHelper, check_call
# import paramiko
import yaml
import pandas as pd

SLURM_PARTITION_RECON = b'slurm_load_partitions: Unable to contact slurm controller (connect failure)'
PARAMIKO_PEM_KEY = os.path.expanduser('~/.ssh/canine_pem_key')
GPU_SCRIPT = ' && '.join([
    'sudo yum install -y kernel-devel-$(uname -r) kernel-headers-$(uname -r)',
    'wget http://developer.download.nvidia.com/compute/cuda/repos/rhel7/x86_64/cuda-repo-rhel7-10.1.168-1.x86_64.rpm',
    'sudo yum install -y cuda-repo-rhel7-10.1.168-1.x86_64.rpm',
    'sudo yum -y updateinfo',
    'sudo yum install -y cuda',
    'wget http://developer.download.nvidia.com/compute/machine-learning/repos/rhel7/x86_64/nvidia-machine-learning-repo-rhel7-1.0.0-1.x86_64.rpm',
    'sudo yum install -y nvidia-machine-learning-repo-rhel7-1.0.0-1.x86_64.rpm',
    'sudo yum -y updateinfo',
    'sudo yum install -y cuda-10-0 libcudnn7 libcudnn7-devel libnvinfer5',
    'curl -s -L https://nvidia.github.io/nvidia-docker/$(. /etc/os-release;echo $ID$VERSION_ID)/nvidia-docker.repo | sudo tee /etc/yum.repos.d/nvidia-docker.repo',
    'sudo yum -y updateinfo',
    'sudo yum install -y nvidia-docker2'
])

class TransientGCPSlurmBackend(RemoteSlurmBackend):
    """
    Backend for transient slurm clusters which need to be deployed and configured
    on GCP before they're ready for use
    """

    def __init__(self, name='slurm-canine', *, max_node_count=10, compute_zone='us-central1-a', controller_type='n1-standard-16', login_type='n1-standard-1', worker_type='n1-highcpu-2', login_count=0, compute_disk_size=20, controller_disk_size=200, gpu_type=None, gpu_count=0, compute_script="", controller_script=""):
        super().__init__('{}-controller.{}.{}'.format(
            name,
            compute_zone,
            get_default_gcp_project()
        ))
        self.config = {
          "cluster_name": name,
          "static_node_count": 0,
          "max_node_count": int(max_node_count),
          "zone": compute_zone,
          "region": compute_zone[:-2],
          "cidr": "10.10.0.0/16",
          "controller_machine_type": controller_type,
          "compute_machine_type": worker_type,
          "compute_disk_type": "pd-standard",
          "compute_disk_size_gb": int(compute_disk_size),
          "controller_disk_type": "pd-ssd",
          "controller_disk_size_gb": int(controller_disk_size),
          "controller_secondary_disk": False,
          "login_machine_type": login_type,
          "login_node_count": int(login_count),
          "login_disk_size_gb": 10,
          "preemptible_bursting": True,
          "private_google_access": True,
          "vpc_net": "default",
          "vpc_subnet": "default",
          "default_users": getpass.getuser(),
          'gpu_count': 0
        }

        if gpu_type is not None and gpu_count > 0:
            self.config['gpu_type'] = gpu_type
            self.config['gpu_count'] = gpu_count

        self.startup_script = """
        sudo yum install -y yum-utils device-mapper-persistent-data lvm2 libcgroup libcgroup-tools htop gcc python-devel python-setuptools redhat-rpm-config wget
        sudo yum-config-manager --add-repo https://download.docker.com/linux/centos/docker-ce.repo
        sudo yum install -y docker-ce
        sudo groupadd docker
        sudo usermod -aG docker {0}
        sudo systemctl enable docker.service
        sudo systemctl start docker.service
        sudo chown root:docker /var/run/docker.sock
        sudo bash -c "echo {0}$'\\t'ALL='(ALL:ALL)'$'\\t'ALL >> /etc/sudoers"
        sudo sed -e 's/GRUB_CMDLINE_LINUX="\\?\\([^"]*\\)"\\?/GRUB_CMDLINE_LINUX="\\1 cgroup_enable=memory swapaccount=1"/' < /etc/default/grub > grub.tmp
        sudo mv grub.tmp /etc/default/grub
        sudo grub2-mkconfig -o /etc/grub2.cfg
        curl https://bootstrap.pypa.io/get-pip.py -o get-pip.py
        sudo python get-pip.py
        sudo pip uninstall -y crcmod
        sudo pip install --no-cache-dir -U crcmod
        {1}
        {2}
        """.format(
            getpass.getuser(),
            GPU_SCRIPT if self.config['gpu_count'] > 0 else '',
            compute_script
        )

        self.controller_script = """
        sudo dd if=/dev/zero of=/swapfile count=4096 bs=1MiB
        sudo chmod 700 /swapfile
        sudo mkswap /swapfile
        sudo swapon /swapfile
        sudo yum install -y gcc python-devel python-setuptools redhat-rpm-config htop
        curl https://bootstrap.pypa.io/get-pip.py -o get-pip.py
        sudo python get-pip.py
        sudo pip uninstall -y crcmod
        sudo pip install --no-cache-dir -U crcmod
        {0}
        """.format(controller_script)

        subprocess.check_call(
            'touch ~/.ssh/google_compute_known_hosts',
            shell=True,
            executable='/bin/bash'
        )

    def __enter__(self):
        """
        Create NFS server
        Set up the NFS server and SLURM cluster
        Use default VPC
        """
        # Idea: Allow the final deployment step to run in a Popen
        # cluster.run then polls the object to make sure the deployment has finished
        # allows copytree to begin while deployment still in progress

        # User may also provide an NFS server IP

        # 1) create NFS: (n1-highcpu-16 [16CPU, 14GB, 4GB SW], SSD w/ user GB)
        # 2) SCP setup script to NFS
        # 3) SSH into NFS, run script
        try:
            with tempfile.TemporaryDirectory() as tempdir:
                shutil.copytree(
                    os.path.join(
                        os.path.dirname(__file__),
                        'slurm-gcp'
                    ),
                    tempdir+'/slurm'
                )
                with open(os.path.join(tempdir, 'slurm', 'scripts', 'custom-compute-install'), 'w') as w:
                    w.write(self.startup_script)
                with open(os.path.join(tempdir, 'slurm', 'scripts', 'custom-controller-install'), 'w') as w:
                    w.write(self.controller_script)
                with open(os.path.join(tempdir, 'slurm', 'slurm-cluster.yaml'), 'w') as w:
                    yaml.dump(
                        {
                            "imports": [
                              {
                                "path": 'slurm.jinja'
                              }
                            ],
                            "resources": [
                              {
                                "name": "slurm-cluster",
                                "type": "slurm.jinja",
                                "properties": self.config
                              }
                            ]
                        },
                        w
                    )
                subprocess.check_call(
                    'gcloud deployment-manager deployments create {} --config {}'.format(
                        self.config['cluster_name'],
                        os.path.join(tempdir, 'slurm', 'slurm-cluster.yaml')
                    ),
                    shell=True,
                    executable='/bin/bash'
                )
            subprocess.check_call(
                'gcloud compute config-ssh',
                shell=True
            )
            subprocess.check_call(
                'gcloud compute os-login ssh-keys add --key-file ~/.ssh/google_compute_engine.pub >/dev/null',
                shell=True
            )
            self.load_config_args()
            time.sleep(30) # Key propagation time
            super().__enter__()
            print("Waiting for slurm to initialize")
            rc, sout, serr = self.invoke("which sinfo")
            while rc != 0:
                time.sleep(10)
                rc, sout, serr = self.invoke("which sinfo")
            time.sleep(60)
            print("Slurm controller is ready. Please call .wait_for_cluster_ready() to wait until the slurm compute nodes are ready to accept work")
            return self
        except:
            self.stop()
            raise

    def stop(self):
        """
        Kills the slurm cluster
        """
        subprocess.check_call(
            'gcloud compute config-ssh --remove',
            shell=True
        )
        subprocess.check_call(
            'echo y | gcloud deployment-manager deployments delete {}'.format(
                self.config['cluster_name']
            ),
            shell=True,
            executable='/bin/bash'
        )
        subprocess.run(
            "echo y | gcloud compute images delete $(gcloud compute images list --filter name:{}| awk 'NR>1 {{print $1}}')".format(
                self.config['cluster_name']
            ),
            shell=True,
            executable='/bin/bash'
        )
        subprocess.run(
            "echo y | gcloud compute instances delete --zone {} $(gcloud compute instances list --filter name:{} | awk 'NR>1 {{print $1}}')".format(
                self.config['zone'],
                self.config['cluster_name']
            ),
            shell=True,
            executable='/bin/bash'
        )

    def __exit__(self, *args):
        """
        kill NFS server
        delete the deployment
        """
        super().__exit__()
        self.stop()

    def sinfo(self, *slurmopts: str, **slurmparams: typing.Any) -> pd.DataFrame:
        """
        Shows the current cluster information
        slurmopts and slurmparams are passed into an ArgumentHelper and unpacked
        as command line arguments
        """
        command = 'sinfo'+ArgumentHelper(*slurmopts, **slurmparams).commandline
        status, stdout, stderr = self.invoke(command)
        if status != 0 and SLURM_PARTITION_RECON in stderr.read():
            print("Transient controller timed out while checking partitions. Allowing one additional retry", file=sys.stderr)
            time.sleep(120)
            status, stdout, stderr = self.invoke(command)
        stderr.seek(0,0)
        check_call(command, status, stdout, stderr)
        df = pd.read_fwf(
            stdout,
            index_col=0
        )
        df.index = df.index.map(str)
        return df


# ====
# Old code to get around openssh issues. May be needed in future if ssh-agent trick doesn't work
# @staticmethod
# def get_paramiko_acceptable_key():
#     """
#     Because, openssh keys are not accepted by paramiko,
#     but for some reason, it's the default format on OSx.
#     Returns the filename of a paramiko-accepted 2048-bit rsa key
#     You'll need to copy it over yourself or add it to os-login
#     """
#     if not os.path.isdir(os.path.dirname(PARAMIKO_PEM_KEY)):
#         os.path.makedirs(os.path.dirname(PARAMIKO_PEM_KEY))
#     if not (os.path.isfile(PARAMIKO_PEM_KEY) and os.path.isfile(PARAMIKO_PEM_KEY+'.pub')):
#         print("Generating new ssh keypair for canine/paramiko")
#         try:
#             subprocess.check_call(
#                 "ssh-keygen -t rsa -b 2048 -f {} -N '' -m PEM".format(PARAMIKO_PEM_KEY),
#                 shell=True
#             )
#         except CalledProcessError:
#             # probably fine. On non-openssh implementations "-m PEM" is not
#             # a valid argument. Try anyways
#             subprocess.check_call(
#                 "ssh-keygen -t rsa -b 2048 -f {} -N ''".format(PARAMIKO_PEM_KEY),
#                 shell=True
#             )
#     # Double-check that the key is acceptable
#     paramiko.RSAKey.from_private_key_file(PARAMIKO_PEM_KEY)
#     return PARAMIKO_PEM_KEY
# subprocess.check_call(
#     'gcloud compute --project {} ssh --zone {} {}-controller -- bash -c \'"if [[ ! -d .ssh ]]; then mkdir .ssh; fi"\''.format(
#         get_default_gcp_project(),
#         self.config['zone'],
#         self.config['cluster_name']
#     ),
#     shell=True,
#     executable='/bin/bash'
# )
# subprocess.check_call(
#     'gcloud compute --project {} ssh --zone {} {}-controller -- bash -c \'"cat - >> .ssh/authorized_keys"\' < {}.pub'.format(
#         get_default_gcp_project(),
#         self.config['zone'],
#         self.config['cluster_name'],
#         self.get_paramiko_acceptable_key()
#     ),
#     shell=True,
#     executable='/bin/bash'
# )
# subprocess.check_call(
#     'gcloud compute --project {} ssh --zone {} {}-controller -- bash -c \'"chmod 600 .ssh/authorized_keys"\''.format(
#         get_default_gcp_project(),
#         self.config['zone'],
#         self.config['cluster_name']
#     ),
#     shell=True,
#     executable='/bin/bash'
# )
# self._RemoteSlurmBackend__sshkwargs['key_filename'] = self.get_paramiko_acceptable_key()



import os
import subprocess
import typing
import shlex
from contextlib import ExitStack
from .base import AbstractLocalizer, PathType, Localization
from ..backends import AbstractSlurmBackend, AbstractTransport
from ..utils import get_default_gcp_project, check_call
from agutil import status_bar

class RemoteLocalizer(AbstractLocalizer):
    """
    Remote Only strategy:
    Staging dir is created remotely and files are copied in
    """
    def __enter__(self):
        """
        Enter localizer context.
        May take any setup action required
        """
        with self.backend.transport() as transport:
            if not transport.isdir(self.environment('controller')['CANINE_ROOT']):
                transport.makedirs(self.environment('controller')['CANINE_ROOT'])
            if not transport.isdir(self.environment('controller')['CANINE_COMMON']):
                transport.makedirs(self.environment('controller')['CANINE_COMMON'])
            if not transport.isdir(self.environment('controller')['CANINE_JOBS']):
                transport.makedirs(self.environment('controller')['CANINE_JOBS'])
            if not transport.isdir(self.environment('controller')['CANINE_OUTPUT']):
                transport.makedirs(self.environment('controller')['CANINE_OUTPUT'])
        return self

    def localize_file(self, src: str, dest: PathType, transport: typing.Optional[AbstractTransport] = None):
        """
        Localizes the given file.
        All files are immediately transferred
        """
        if src.startswith('gs://'):
            self.gs_copy(
                src,
                dest.controllerpath,
                'remote'
            )
        elif os.path.exists(src):
            with self.transport_context(transport) as transport:
                if not transport.isdir(os.path.dirname(dest.controllerpath)):
                    transport.makedirs(os.path.dirname(dest.controllerpath))
                if os.path.isfile(src):
                    transport.send(src, dest.controllerpath)
                else:
                    self.sendtree(
                        src,
                        dest.controllerpath,
                        transport
                    )

    def localize(self, inputs: typing.Dict[str, typing.Dict[str, str]], patterns: typing.Dict[str, str], overrides: typing.Optional[typing.Dict[str, typing.Optional[str]]] = None) -> str:
        """
        3 phase task:
        1) Pre-scan inputs to determine proper localization strategy for all inputs
        2) Begin localizing job inputs. For each job, check the predetermined strategy
        and set up the job's setup and teardown scripts
        3) Finally, finalize the localization. This may include broadcasting the
        staging directory or copying a batch of gsutil files
        Returns the remote staging directory, which is now ready for final startup
        """
        if overrides is None:
            overrides = {}
        overrides = {k:v.lower() if isinstance(v, str) else None for k,v in overrides.items()}
        with self.backend.transport() as transport:
            if self.common:
                common_dests = self.pick_common_inputs(inputs, overrides, transport=transport)
            else:
                common_dests = {}
            for jobId, data in inputs.items():
                transport.makedirs(
                    os.path.join(
                        self.environment('controller')['CANINE_JOBS'],
                        jobId
                    )
                )
                self.prepare_job_inputs(jobId, data, common_dests, overrides, transport=transport)
                # Now localize job setup and teardown scripts
                setup_script, teardown_script = self.job_setup_teardown(jobId, patterns)
                # Setup:
                script_path = self.reserve_path('jobs', jobId, 'setup.sh')
                with transport.open(script_path.controllerpath, 'w') as w:
                    w.write(setup_script)
                transport.chmod(script_path.controllerpath, 0o775)
                # Teardown:
                script_path = self.reserve_path('jobs', jobId, 'teardown.sh')
                with transport.open(script_path.controllerpath, 'w') as w:
                    w.write(teardown_script)
                transport.chmod(script_path.controllerpath, 0o775)
            transport.send(
                os.path.join(
                    os.path.dirname(__file__),
                    'delocalization.py'
                ),
                os.path.join(self.environment('controller')['CANINE_ROOT'], 'delocalization.py')
            )
            return self.finalize_staging_dir(inputs.keys(), transport=transport)



import abc
import typing
import re
import shutil
import os
import time
import stat
from contextlib import ExitStack
from uuid import uuid4 as uuid
from ..utils import ArgumentHelper, check_call
import pandas as pd

batch_job_pattern = re.compile(r'Submitted batch job (\d+)')

class AbstractTransport(abc.ABC):
    """
    Base class for file transport
    """
    @abc.abstractmethod
    def __enter__(self):
        """
        Allows the Transport to function as a context manager
        No action is required, so long as the context meets these requirements
        * Any required initialization takes place during __enter__
        * All transport methods work within the context
        * If initialization is required, transport methods fail outside context
        """
        pass

    @abc.abstractmethod
    def __exit__(self, *args):
        """
        Allows the Transport to function as a context manager
        No action is required, so long as the context meets these requirements
        * Any required teardown takes place during __exit__
        * If teardown is required, transport methods fail outside context
        """
        pass

    @abc.abstractmethod
    def open(self, filename: str, mode: str = 'r', bufsize: int = -1) -> typing.IO:
        """
        Returns a File-Like object open on the slurm cluster
        """
        pass

    def send(self, localfile: typing.Union[str, typing.IO], remotefile: typing.Union[str, typing.IO]):
        """
        Sends the requested file to the slurm cluster
        Both localfile and remotefile can either be a filepath or a file object
        If remotefile is a file object, it must have been opened by this transport
        or the copy will not behave as expected
        """
        with ExitStack() as stack:
            if isinstance(localfile, str):
                localfile = stack.enter_context(open(localfile, 'rb'))
            if isinstance(remotefile, str):
                remotefile = stack.enter_context(self.open(remotefile, 'wb' if 'b' in localfile.mode else 'w'))
            shutil.copyfileobj(localfile, remotefile)
            self.chmod(remotefile.name, os.stat(localfile.name).st_mode)

    def receive(self, remotefile: typing.Union[str, typing.IO], localfile: typing.Union[str, typing.IO]):
        """
        Copies the requested file from the slurm cluster
        Both localfile and remotefile can either be a filepath or a file object
        If remotefile is a file object, it must have been opened by this transport
        or the copy will not behave as expected
        """
        with ExitStack() as stack:
            if isinstance(remotefile, str):
                remotefile = stack.enter_context(self.open(remotefile, 'rb'))
            if isinstance(localfile, str):
                localfile = stack.enter_context(open(localfile, 'wb' if 'b' in remotefile.mode else 'w'))
            shutil.copyfileobj(remotefile, localfile)
            os.chmod(localfile.name, self.stat(remotefile.name).st_mode)

    @abc.abstractmethod
    def listdir(self, path: str) -> typing.List[str]:
        """
        Lists the contents of the requested path
        """
        pass

    @abc.abstractmethod
    def mkdir(self, path: str):
        """
        Creates the requested directory
        """
        pass

    @abc.abstractmethod
    def stat(self, path: str) -> typing.Any:
        """
        Returns stat information
        """
        pass

    def exists(self, path: str) -> bool:
        """
        Returns True if the given path exists
        """
        try:
            self.stat(path)
            return True
        except FileNotFoundError:
            return False

    @abc.abstractmethod
    def chmod(self, path: str, mode: int):
        """
        Change file permissions
        """
        pass

    @abc.abstractmethod
    def normpath(self, path: str) -> str:
        """
        Returns a normalized path relative to the transport
        """
        pass

    @abc.abstractmethod
    def remove(self, path: str):
        """
        Removes the file at the given path
        """
        pass

    @abc.abstractmethod
    def rmdir(self, path: str):
        """
        Removes the directory at the given path
        """
        pass

    @abc.abstractmethod
    def rename(self, src: str, dest: str):
        """
        Move the file or folder 'src' to 'dest'
        Will overwrite dest if it exists
        """
        pass

    @abc.abstractmethod
    def mklink(self, src: str, dest: str):
        """
        Creates a symlink from dest->src
        """
        pass

    def isdir(self, path: str) -> bool:
        """
        Returns True if the requested path is a directory
        """
        try:
            return stat.S_ISDIR(self.stat(path).st_mode)
        except FileNotFoundError:
            return False

    def isfile(self, path: str) -> bool:
        """
        Returns True if the requested path is a regular file
        """
        try:
            return stat.S_ISREG(self.stat(path).st_mode)
        except FileNotFoundError:
            return False

    def islink(self, path: str) -> bool:
        """
        Returns True if the requested path is a symlink
        """
        try:
            return stat.S_ISLNK(self.stat(path).st_mode)
        except FileNotFoundError:
            return False

    def makedirs(self, path: str):
        """
        Recursively build the requested directory structure
        """
        if path == '' or self.isdir(path):
            return
        dirname = os.path.dirname(path)
        if not (dirname == '' or os.path.exists(dirname)):
            self.makedirs(dirname)
        self.mkdir(path)

    def walk(self, path: str) -> typing.Generator[typing.Tuple[str, typing.List[str], typing.List[str]], None, None]:
        """
        Walk through a directory tree
        Each iteration yields a 3-tuple:
        (dirpath, dirnames, filenames) ->
        * dirpath: The current filepath relative to the starting path
        * dirnames: The base names of all subdirectories in the current directory
        * filenames: The base names of all files in the current directory
        """
        dirnames = []
        filenames = []
        for name in self.listdir(path):
            if self.isdir(os.path.join(path, name)):
                dirnames.append(name)
            else:
                filenames.append(name)
        yield (path, dirnames, filenames)
        for dirname in dirnames:
            yield from self.walk(os.path.join(path, dirname))

    def sendtree(self, src: str, dest: str):
        """
        Copy the full local file tree src to the remote path dest
        """
        if not self.exists(dest):
            self.makedirs(dest)
        for path, dirnames, filenames in os.walk(src):
            rpath = os.path.join(
                dest,
                os.path.relpath(path, src)
            )
            if not self.exists(rpath):
                self.mkdir(rpath)
            for f in filenames:
                self.send(os.path.join(path, f), os.path.join(rpath, f))


    def receivetree(self, src: str, dest: str):
        """
        Copy the full remote file tree src to the local path dest
        """
        if not os.path.exists(dest):
            os.makedirs(dest)
        for path, dirnames, filenames in self.walk(src):
            lpath = os.path.join(
                dest,
                os.path.relpath(path, src)
            )
            if not os.path.exists(lpath):
                os.mkdir(lpath)
            for f in filenames:
                self.receive(os.path.join(path, f), os.path.join(lpath, f))

class AbstractSlurmBackend(abc.ABC):
    """
    Base class for a SLURM backend
    """

    @abc.abstractmethod
    def invoke(self, command: str, interactive: bool = False) -> typing.Tuple[int, typing.BinaryIO, typing.BinaryIO]:
        """
        Invoke an arbitrary command in the slurm console
        Returns a tuple containing (exit status, byte stream of standard out from the command, byte stream of stderr from the command).
        If interactive is True, stdin, stdout, and stderr should all be connected live to the user's terminal
        """
        pass

    def squeue(self, *slurmopts: str, **slurmparams: typing.Any) -> pd.DataFrame:
        """
        Shows the current status of the job queue
        slurmopts and kwargs are passed into an ArgumentHelper and unpacked
        as command line arguments
        """
        command = 'squeue'+ArgumentHelper(*slurmopts, **slurmparams).commandline
        status, stdout, stderr = self.invoke(command)
        check_call(command, status, stdout, stderr)
        df = pd.read_fwf(
            stdout,
            index_col=0
        )
        df.index = df.index.map(str)
        return df


    def sacct(self, *slurmopts: str, **slurmparams: typing.Any) -> pd.DataFrame:
        """
        Shows the current job accounting information
        slurmopts and slurmparams are passed into an ArgumentHelper and unpacked
        as command line arguments
        """
        command = 'sacct'+ArgumentHelper(*slurmopts, **slurmparams).commandline
        status, stdout, stderr = self.invoke(command)
        check_call(command, status, stdout, stderr)
        df = pd.read_fwf(
            stdout,
            index_col=0
        )
        df.index = df.index.map(str)
        return df.loc[[idx for idx in df.index if not idx.startswith('---')]]

    def sinfo(self, *slurmopts: str, **slurmparams: typing.Any) -> pd.DataFrame:
        """
        Shows the current cluster information
        slurmopts and slurmparams are passed into an ArgumentHelper and unpacked
        as command line arguments
        """
        command = 'sinfo'+ArgumentHelper(*slurmopts, **slurmparams).commandline
        status, stdout, stderr = self.invoke(command)
        check_call(command, status, stdout, stderr)
        df = pd.read_fwf(
            stdout,
            index_col=0
        )
        df.index = df.index.map(str)
        return df

    def srun(self, command: str, *slurmopts: str, **slurmparams: typing.Any) -> typing.Tuple[int, typing.BinaryIO, typing.BinaryIO]:
        """
        Runs the given command interactively
        The first argument MUST contain the entire command and options to run
        Additional arguments and keyword arguments provided are passed as slurm arguments to srun
        Returns a tuple containing (exit status, stdout buffer, stderr buffer)
        """
        command = 'srun {} -- {}'.format(
            ArgumentHelper(*slurmopts, **slurmparams).commandline,
            command
        )
        status, stdout, stderr = self.invoke(command, True)
        check_call(command, status) # Don't pass stdout and stderr here, they're already printed live to terminal
        return status, stdout, stderr

    def sbatch(self, command: str, *slurmopts: str, **slurmparams: typing.Any) -> str:
        """
        Runs the given command in the background
        The first argument MUST contain the entire command and options to run
        Additional arguments and keyword arguments provided are passed as slurm arguments to sbatch
        Returns the jobID of the batch request
        """
        command = 'sbatch {} -- {}'.format(
            ArgumentHelper(*slurmopts, **slurmparams).commandline,
            command
        )
        status, stdout, stderr = self.invoke(command)
        check_call(command, status, stdout, stderr)
        result = batch_job_pattern.search(stdout.read().decode())
        if result:
            return result.group(1)
        raise ValueError("Command output did not match pattern")

    def scancel(self, jobID: str, *slurmopts: str, **slurmparams: typing.Any):
        """
        Cancels the given jobID
        The first argument MUST be the jobID
        Additional arguments and keyword arguments are passed as slurm arguments
        to scancel
        """
        command = 'scancel {} {}'.format(
            ArgumentHelper(*slurmopts, **slurmparams).commandline,
            jobID
        )
        status, stdout, stderr = self.invoke(command)
        check_call(command, status, stdout, stderr)


    @abc.abstractmethod
    def __enter__(self):
        """
        Allows backends to serve as context managers
        The context is not explicitly required to do anything, but must meet the
        following:
        * Any required initialization happens during __enter__
        * Slurm commands will be connected and functional inside the context
        * If the backend requires an initialization step, slurm commands will not
        function before entering the context
        """
        pass

    @abc.abstractmethod
    def __exit__(self, *args):
        """
        Allows backends to serve as context managers
        The context is not explicitly required to do anything, but must meet the
        following:
        * Any required teardown happens during __exit__
        * If the backend requires explicit teardown, slurm commands will not function
        after exiting the context
        """
        pass

    @abc.abstractmethod
    def transport(self) -> AbstractTransport:
        """
        Return a Transport object suitable for moving files between the
        SLURM cluster and the local filesystem
        """
        pass

    def pack_batch_script(self, *commands: str, script_path: str = None):
        """
        writes the list of commands to a file and runs it
        Script_path may be the path to where the script should be written or None
        """
        if script_path is None:
            script_path = str(uuid())+'.sh'
        with self.transport() as transport:
            transport.makedirs(os.path.dirname(script_path))
            with transport.open(script_path, 'w') as w:
                w.write('#!/bin/bash\n')
                for command in commands:
                    w.write(command.rstrip()+'\n')
            transport.chmod(script_path, 0o775)
        return script_path

    def wait_for_cluster_ready(self):
        """
        Blocks until the main partition is marked as up
        """
        df = self.sinfo()
        default = [value for value in df.index if value.endswith('*')]
        while len(default) == 0:
            time.sleep(10)
            df = self.sinfo()
            default = [value for value in df.index if value.endswith('*')]
        while (df.AVAIL[default] != 'up').all():
            time.sleep(10)
            df = self.sinfo()



"""
For argument parsing and CLI interface
"""
import argparse
import sys
from . import Orchestrator
from .backends import TransientGCPSlurmBackend
import yaml

def ConfType(nfields, nMax=None):
    """
    Returns an argument handler which expects the given number of fields
    Less than the specified number results in an error
    More than the specified number results in the extra fields being combined together
    """
    if nMax is None:
        nMax=nfields

    def parse_arg(argument):
        args = argument.split(':')
        if len(args) < nfields:
            raise argparse.ArgumentError(argument, "Not enough fields in argument (expected {}, got {})".format(nfields, len(args)))
        elif len(args) > nMax:
            args = args[:nMax-1] + ':'.join(args[nMax-1:])
        return args

    return parse_arg

def boot_transient():
    parser = argparse.ArgumentParser(
        'canine-transient',
        description="Boots a transient slurm cluster"
    )
    parser.add_argument(
        'name',
        help="Name of the cluster"
    )
    parser.add_argument(
        '-n', '--node-count',
        type=int,
        help="Max number of compute nodes in the cluster",
        default=10
    )
    parser.add_argument(
        '-z', '--zone',
        help="GCP Compute zone in which to create the cluster",
        default='us-central1-a'
    )
    parser.add_argument(
        '-c', '--controller-type',
        help="Instance type for controller node",
        default="n1-standard-16"
    )
    parser.add_argument(
        '-w', '--worker-type',
        help="Instance type for compute node",
        default="n1-highcpu-2"
    )
    parser.add_argument(
        '-s', '--controller-disk-size',
        type=int,
        help="Size of the controller's disk in GB",
        default=200
    )
    parser.add_argument(
        '-d', '--worker-disk-size',
        type=int,
        help="Size of the compute node disk in GB",
        default=20
    )
    parser.add_argument(
        '-t', '--gpu-type',
        help="Type of GPU to attach to compute nodes",
        default=None
    )
    parser.add_argument(
        '-g', '--gpu-count',
        type=int,
        help="Number of GPUs to attach to compute nodes",
        default=0
    )
    args = parser.parse_args()
    with TransientGCPSlurmBackend(
        name=args.name,
        max_node_count=args.node_count,
        compute_zone=args.zone,
        controller_type=args.controller_type,
        worker_type=args.worker_type,
        compute_disk_size=args.worker_disk_size,
        controller_disk_size=args.controller_disk_size,
        gpu_type=args.gpu_type,
        gpu_count=args.gpu_count
    ) as slurm:
        print("=====================")
        print("Slurm cluster started")
        print("SSH using", slurm._RemoteSlurmBackend__hostname)
        print("=====================")
        try:
            while True:
                input("Press Ctrl+C to kill the slurm cluster")
        except KeyboardInterrupt:
            pass # silence keyboard interrupt

def main():
    parser = argparse.ArgumentParser(
        'canine',
        description="A dalmatian-based job manager to schedule tasks using SLURM"
    )
    parser.add_argument(
        'pipeline',
        nargs='?',
        type=argparse.FileType('r'),
        help="Path to a pipeline file. Command line arguments will merge with,"
        "and override options in the file",
        default=None
    )
    parser.add_argument(
        '--dry-run',
        action='store_true',
        help="If provided, the job will not actually run. Canine will parse job"
        " inputs and walk through localization, but will not ever schedule the job."
        " All inputs and job scripts will be prepared and localized in the staging"
        " directory"
    )
    parser.add_argument(
        '--export',
        type=argparse.FileType('w'),
        help="If provided, Canine will write the final merged pipeline object to"
        " the provided filepath",
        default=None
    )
    parser.add_argument(
        '--output-dir',
        help="Output directory for canine pipeline. Defaults to 'canine_output'",
        default=None
    )
    parser.add_argument(
        '-n', '--name',
        help="Name of the job",
        default=None
    )
    parser.add_argument(
        '-s', '--script',
        help="Path to the script to run",
        default=None,
        type=argparse.FileType('r')
    ),
    parser.add_argument(
        '-i', '--input',
        help="Script inputs. Must specify in the form inputName:inputValue. --input"
        " may be specified as many times as necessary, and inputNames may also be repeated",
        type=ConfType(2),
        action='append',
        default=[]
    )
    parser.add_argument(
        '-r', '--resources',
        help="SLURM arguments. Must specify in the form argName:argValue. --resources"
        " may be specified as many times as necessary, and a specific argName may also"
        " be repeated. Specify slurm arguments without leading dashes, but otherwise"
        " exactly as they'd appear on the command line. For slurm options"
        " which take no arguments, set --resources argName:true",
        type=ConfType(2),
        action='append',
        default=[]
    )
    parser.add_argument(
        '-a', '--adapter',
        help="Adapter configuration. Must specify in the form optionName:optionValue."
        " --adapter may be specified as many times as necessary",
        type=ConfType(2),
        action='append',
        default=[]
    )
    parser.add_argument(
        '-b', '--backend',
        help="Backend configuration. Must specify in the form optionName:optionValue."
        " --backend may be provided as many times as necessary",
        type=ConfType(2),
        action='append',
        default=[]
    )
    parser.add_argument(
        '-l', '--localization',
        help="Localization configuration. Must specify in the form optionName:optionValue."
        " --localization may be provided as many times as necessary. localization"
        " overrides should be specified using --localization overrides:outputName:overrideValue",
        type=ConfType(2,3),
        action='append',
        default=[]
    )
    parser.add_argument(
        '-o', '--output',
        help="Output patterns. Must specify in the form outputName:globPattern."
        " --output may be provided as many times as necessary.",
        type=ConfType(2),
        action='append',
        default=[]
    )
    args = parser.parse_args()
    conf = {}
    if args.pipeline:
        conf = yaml.load(args.pipeline, Loader=yaml.loader.SafeLoader)
    if args.name is not None:
        conf['name'] = args.name
    if args.script is not None:
        conf['script'] = args.script.name
    if len(args.resources) > 0:
        if 'resources' not in conf:
            conf['resources'] = {}
        conf['resources'] = {
            **conf['resources'],
            **{key: val for key, val in args.resources}
        }
    if len(args.adapter) > 0:
        if 'adapter' not in conf:
            conf['adapter'] = {}
        conf['adapter'] = {
            **conf['adapter'],
            **{key: val for key, val in args.adapter}
        }
    if len(args.backend) > 0:
        if 'backend' not in conf:
            conf['backend'] = {}
        conf['backend'] = {
            **conf['backend'],
            **{key: val for key, val in args.backend}
        }
    if len(args.output) > 0:
        if 'outputs' not in conf:
            conf['outputs'] = {}
            conf['outputs'] = {
                **conf['outputs'],
                **{key: val for key, val in args.output}
            }
    if len(args.input) > 0:
        inputs = {}
        for name, val in args.input:
            if name in inputs:
                if isinstance(inputs[name], list):
                    inputs[name].append(val)
                else:
                    inputs[name] = [inputs[name], val]
            else:
                inputs[name] = val
        if 'inputs' not in conf:
            conf['inputs'] = {}
        conf['inputs'] = {
            **conf['inputs'],
            **inputs
        }
    if len(args.localization) > 0:
        overrides = {}
        localization = {}
        for entry in args.localization:
            if entry[0] == 'overrides':
                overrides[entry[1]] = entry[2]
            else:
                localization[entry[0]] = entry[1]
        if 'localization' not in conf:
            conf['localization'] = {'overrides':{}}
        conf['localization'] = {
            **conf['localization'],
            **localization
        }
        if len(overrides):
            if 'overrides' not in conf['localization']:
                conf['localization']['overrides'] = {}
            conf['localization']['overrides'] = {
                **conf['localization']['overrides'],
                **overrides
            }
    if args.export is not None:
        yaml.dump(conf, args.export)
    if not len(conf):
        sys.exit("Empty pipeline config")
    kwargs = {'dry_run': args.dry_run}
    if args.output_dir is not None:
        kwargs['output_dir'] = args.output_dir
    Orchestrator(conf).run_pipeline(**kwargs)

if __name__ == '__main__':
    main()
